# encoding: utf-8
from Aspidites._vendor.contracts import contract
from Aspidites._vendor.contracts.main import parse_contract_string
from Aspidites._vendor.contracts.library import Extension
name = 'helló wörld from one'

import unittest

class TestParsingNumbers(unittest.TestCase):

    def test_unicode_literal(self):
        r = parse_contract_string(u'int')
        assert r == Extension('int')

    def test_unicode_literal2(self):

        @contract(x='string')
        def f(x):
            pass


        f('')



from .suggester import create_suggester
from .dummy import Any, Never
from .separate_context import SeparateContext
from .types_misc import Type, CheckType, Number
from .strings import *
from .lists import List
from .seq import Seq
from .tuple import Tuple
from .collection import Collection
from .datetime_tz import DatetimeWithTz
from .dicts import Dict
from .map import Map
from .sets import *
from .attributes import Attr
from .files import File


from .comparison import CheckOrder
from .arithmetic import Binary, Unary
from .compositions import OR, And, Not, composite_contract, or_contract
from .variables import (BindVariable, VariableRef, misc_variables_contract,
                        int_variables_contract, misc_variables_ref,
                        int_variables_ref)
from .simple_values import EqualTo, SimpleRValue

try:
    import numpy
except ImportError:  # pragma: no cover
    pass
else:
    from .array import (ShapeContract, Shape, Array, ArrayConstraint, DType,
                        dtype, ArrayOR, ArrayAnd)

from .extensions import (identifier_expression, identifier_contract,
                         Extension, CheckCallable)

from .isinstance_imp import *
from .scoped_variables import scoped_variables_ref

from .miscellaneous_aliases import *


# -*- coding: utf-8 -*-
"""\

The :class:`set` type brings the practical expressiveness of
set theory to Python. It has a very rich API overall, but lacks a
couple of fundamental features. For one, sets are not ordered. On top
of this, sets are not indexable, i.e, ``my_set[8]`` will raise an
:exc:`TypeError`. The :class:`IndexedSet` type remedies both of these
issues without compromising on the excellent complexity
characteristics of Python's built-in set implementation.
"""

from __future__ import print_function

from bisect import bisect_left
from itertools import chain, islice
import operator

try:
    from collections.abc import MutableSet
except ImportError:
    from collections import MutableSet

try:
    from typeutils import make_sentinel

    _MISSING = make_sentinel(var_name='_MISSING')
except ImportError:
    _MISSING = object()

__all__ = ['complement']

_COMPACTION_FACTOR = 8


# TODO: inherit from set()
# TODO: .discard_many(), .remove_many()
# TODO: raise exception on non-set params?
# TODO: technically reverse operators should probably reverse the
# order of the 'other' inputs and put self last (to try and maintain
# insertion order)


def complement(wrapped):
    """Given a :class:`set`, convert it to a **complement set**.

    Whereas a :class:`set` keeps track of what it contains, a
    `complement set
    <https://en.wikipedia.org/wiki/Complement_(set_theory)>`_ keeps
    track of what it does *not* contain. For example, look what
    happens when we intersect a normal set with a complement set::

    >>> list(set(range(5)) & complement(set([2, 3])))
    [0, 1, 4]

    We get the everything in the left that wasn't in the right,
    because intersecting with a complement is the same as subtracting
    a normal set.

    Args:
        wrapped (set): A set or any other iterable which should be
           turned into a complement set.

    All set methods and operators are supported by complement sets,
    between other :func:`complement`-wrapped sets and/or regular
    :class:`set` objects.

    Because a complement set only tracks what elements are *not* in
    the set, functionality based on set contents is unavailable:
    :func:`len`, :func:`iter` (and for loops), and ``.pop()``. But a
    complement set can always be turned back into a regular set by
    complementing it again:

    >>> s = set(range(5))
    >>> complement(complement(s)) == s
    True

    .. note::

       An empty complement set corresponds to the concept of a
       `universal set <https://en.wikipedia.org/wiki/Universal_set>`_
       from mathematics.

    Complement sets by example
    ^^^^^^^^^^^^^^^^^^^^^^^^^^

    Many uses of sets can be expressed more simply by using a
    complement. Rather than trying to work out in your head the proper
    way to invert an expression, you can just throw a complement on
    the set. Consider this example of a name filter::

        >>> class NamesFilter(object):
        ...    def __init__(self, allowed):
        ...        self._allowed = allowed
        ...
        ...    def filter(self, names):
        ...        return [name for name in names if name in self._allowed]
        >>> NamesFilter(set(['alice', 'bob'])).filter(['alice', 'bob', 'carol'])
        ['alice', 'bob']

    What if we want to just express "let all the names through"?

    We could try to enumerate all of the expected names::

       ``NamesFilter({'alice', 'bob', 'carol'})``

    But this is very brittle -- what if at some point over this
    object is changed to filter ``['alice', 'bob', 'carol', 'dan']``?

    Even worse, what about the poor programmer who next works
    on this piece of code?  They cannot tell whether the purpose
    of the large allowed set was "allow everything", or if 'dan'
    was excluded for some subtle reason.

    A complement set lets the programmer intention be expressed
    succinctly and directly::

       NamesFilter(complement(set()))

    Not only is this code short and robust, it is easy to understand
    the intention.

    """
    if type(wrapped) is _ComplementSet:
        return wrapped.complemented()
    if type(wrapped) is frozenset:
        return _ComplementSet(excluded=wrapped)
    return _ComplementSet(excluded=set(wrapped))


def _norm_args_typeerror(other):
    '''normalize args and raise type-error if there is a problem'''
    if type(other) in (set, frozenset):
        inc, exc = other, None
    elif type(other) is _ComplementSet:
        inc, exc = other._included, other._excluded
    else:
        raise TypeError('argument must be another set or complement(set)')
    return inc, exc


def _norm_args_notimplemented(other):
    '''normalize args and return NotImplemented (for overloaded operators)'''
    if type(other) in (set, frozenset):
        inc, exc = other, None
    elif type(other) is _ComplementSet:
        inc, exc = other._included, other._excluded
    else:
        return NotImplemented, None
    return inc, exc


class _ComplementSet(object):
    """
    helper class for complement() that implements the set methods
    """
    __slots__ = ('_included', '_excluded')

    def __init__(self, included=None, excluded=None):
        if included is None:
            assert type(excluded) in (set, frozenset)
        elif excluded is None:
            assert type(included) in (set, frozenset)
        else:
            raise ValueError('one of included or excluded must be a set')
        self._included, self._excluded = included, excluded

    def __repr__(self):
        if self._included is None:
            return 'complement({0})'.format(repr(self._excluded))
        return 'complement(complement({0}))'.format(repr(self._included))

    def complemented(self):
        '''return a complement of the current set'''
        if type(self._included) is frozenset or type(self._excluded) is frozenset:
            return _ComplementSet(included=self._excluded, excluded=self._included)
        return _ComplementSet(
            included=None if self._excluded is None else set(self._excluded),
            excluded=None if self._included is None else set(self._included))

    __invert__ = complemented

    def complement(self):
        '''convert the current set to its complement in-place'''
        self._included, self._excluded = self._excluded, self._included

    def __contains__(self, item):
        if self._included is None:
            return not item in self._excluded
        return item in self._included

    def add(self, item):
        if self._included is None:
            if item in self._excluded:
                self._excluded.remove(item)
        else:
            self._included.add(item)

    def remove(self, item):
        if self._included is None:
            self._excluded.add(item)
        else:
            self._included.remove(item)

    def pop(self):
        if self._included is None:
            raise NotImplementedError  # self.missing.add(random.choice(gc.objects()))
        return self._included.pop()

    def intersection(self, other):
        try:
            return self & other
        except NotImplementedError:
            raise TypeError('argument must be another set or complement(set)')

    def __and__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return _ComplementSet(included=inc - self._excluded)
            else:  # - -
                return _ComplementSet(excluded=self._excluded.union(other._excluded))
        else:
            if inc is None:  # + -
                return _ComplementSet(included=exc - self._included)
            else:  # + +
                return _ComplementSet(included=self._included.intersection(inc))

    __rand__ = __and__

    def __iand__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                self._excluded = inc - self._excluded  # TODO: do this in place?
            else:  # - -
                self._excluded |= exc
        else:
            if inc is None:  # + -
                self._included -= exc
                self._included, self._excluded = None, self._included
            else:  # + +
                self._included &= inc
        return self

    def union(self, other):
        try:
            return self | other
        except NotImplementedError:
            raise TypeError('argument must be another set or complement(set)')

    def __or__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return _ComplementSet(excluded=self._excluded - inc)
            else:  # - -
                return _ComplementSet(excluded=self._excluded.intersection(exc))
        else:
            if inc is None:  # + -
                return _ComplementSet(excluded=exc - self._included)
            else:  # + +
                return _ComplementSet(included=self._included.union(inc))

    __ror__ = __or__

    def __ior__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                self._excluded -= inc
            else:  # - -
                self._excluded &= exc
        else:
            if inc is None:  # + -
                self._included, self._excluded = None, exc - self._included  # TODO: do this in place?
            else:  # + +
                self._included |= inc
        return self

    def update(self, items):
        if type(items) in (set, frozenset):
            inc, exc = items, None
        elif type(items) is _ComplementSet:
            inc, exc = items._included, items._excluded
        else:
            inc, exc = frozenset(items), None
        if self._included is None:
            if exc is None:  # - +
                self._excluded &= inc
            else:  # - -
                self._excluded.discard(exc)
        else:
            if inc is None:  # + -
                self._included &= exc
                self._included, self._excluded = None, self._excluded
            else:  # + +
                self._included.update(inc)

    def discard(self, items):
        if type(items) in (set, frozenset):
            inc, exc = items, None
        elif type(items) is _ComplementSet:
            inc, exc = items._included, items._excluded
        else:
            inc, exc = frozenset(items), None
        if self._included is None:
            if exc is None:  # - +
                self._excluded.update(inc)
            else:  # - -
                self._included, self._excluded = exc - self._excluded, None
        else:
            if inc is None:  # + -
                self._included &= exc
            else:  # + +
                self._included.discard(inc)

    def symmetric_difference(self, other):
        try:
            return self ^ other
        except NotImplementedError:
            raise TypeError('argument must be another set or complement(set)')

    def __xor__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return _ComplementSet(excluded=self._excluded - inc)
            else:  # - -
                return _ComplementSet(included=self._excluded.symmetric_difference(exc))
        else:
            if inc is None:  # + -
                return _ComplementSet(excluded=exc - self._included)
            else:  # + +
                return _ComplementSet(included=self._included.symmetric_difference(inc))

    __rxor__ = __xor__

    def symmetric_difference_update(self, other):
        inc, exc = _norm_args_typeerror(other)
        if self._included is None:
            if exc is None:  # - +
                self._excluded |= inc
            else:  # - -
                self._excluded.symmetric_difference_update(exc)
                self._included, self._excluded = self._excluded, None
        else:
            if inc is None:  # + -
                self._included |= exc
                self._included, self._excluded = None, self._included
            else:  # + +
                self._included.symmetric_difference_update(inc)

    def isdisjoint(self, other):
        inc, exc = _norm_args_typeerror(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return inc.issubset(self._excluded)
            else:  # - -
                return False
        else:
            if inc is None:  # + -
                return self._included.issubset(exc)
            else:  # + +
                return self._included.isdisjoint(inc)

    def issubset(self, other):
        '''everything missing from other is also missing from self'''
        try:
            return self <= other
        except NotImplementedError:
            raise TypeError('argument must be another set or complement(set)')

    def __le__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return False
            else:  # - -
                return self._excluded.issupserset(exc)
        else:
            if inc is None:  # + -
                return self._included.isdisjoint(exc)
            else:  # + +
                return self._included.issubset(inc)

    def __lt__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return False
            else:  # - -
                return self._excluded > exc
        else:
            if inc is None:  # + -
                return self._included.isdisjoint(exc)
            else:  # + +
                return self._included < inc

    def issuperset(self, other):
        '''everything missing from self is also missing from super'''
        try:
            return self >= other
        except NotImplementedError:
            raise TypeError('argument must be another set or complement(set)')

    def __ge__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return not self._excluded.intersection(inc)
            else:  # - -
                return self._excluded.issubset(exc)
        else:
            if inc is None:  # + -
                return False
            else:  # + +
                return self._included.issupserset(inc)

    def __gt__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return not self._excluded.intersection(inc)
            else:  # - -
                return self._excluded < exc
        else:
            if inc is None:  # + -
                return False
            else:  # + +
                return self._included > inc

    def difference(self, other):
        try:
            return self - other
        except NotImplementedError:
            raise TypeError('argument must be another set or complement(set)')

    def __sub__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                return _ComplementSet(excluded=self._excluded | inc)
            else:  # - -
                return _ComplementSet(included=exc - self._excluded)
        else:
            if inc is None:  # + -
                return _ComplementSet(included=self._included & exc)
            else:  # + +
                return _ComplementSet(included=self._included.difference(inc))

    def __rsub__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        # rsub, so the expression being evaluated is "other - self"
        if self._included is None:
            if exc is None:  # - +
                return _ComplementSet(included=inc & self._excluded)
            else:  # - -
                return _ComplementSet(included=self._excluded - exc)
        else:
            if inc is None:  # + -
                return _ComplementSet(excluded=exc | self._included)
            else:  # + +
                return _ComplementSet(included=inc.difference(self._included))

    def difference_update(self, other):
        try:
            self -= other
        except NotImplementedError:
            raise TypeError('argument must be another set or complement(set)')

    def __isub__(self, other):
        inc, exc = _norm_args_notimplemented(other)
        if inc is NotImplemented:
            return NotImplemented
        if self._included is None:
            if exc is None:  # - +
                self._excluded |= inc
            else:  # - -
                self._included, self._excluded = exc - self._excluded, None
        else:
            if inc is None:  # + -
                self._included &= exc
            else:  # + +
                self._included.difference_update(inc)
        return self

    def __eq__(self, other):
        return (
                       type(self) is type(other)
                       and self._included == other._included
                       and self._excluded == other._excluded) or (
                       type(other) in (set, frozenset) and self._included == other)

    def __hash__(self):
        return hash(self._included) ^ hash(self._excluded)

    def __len__(self):
        if self._included is not None:
            return len(self._included)
        raise NotImplementedError('complemented sets have undefined length')

    def __iter__(self):
        if self._included is not None:
            return iter(self._included)
        raise NotImplementedError('complemented sets have undefined contents')

    def __bool__(self):
        if self._included is not None:
            return bool(self._included)
        return True

    __nonzero__ = __bool__  # py2 compat



from numpy import zeros, ones
import numpy as np

from . import syntax_fail, good, fail
from Aspidites._vendor.contracts.library.array import np_int_dtypes, np_uint_dtypes, \
    np_float_dtypes

a_u8 = np.zeros((3, 4), dtype='uint8')
a_i8 = np.zeros((3, 4), dtype='int8')
a_u16 = np.zeros((3, 4), dtype='uint16')
a_i16 = np.zeros((3, 4), dtype='int16')
a_u32 = np.zeros((3, 4), dtype='uint32')
a_i32 = np.zeros((3, 4), dtype='int32')
a_f32 = np.zeros((3, 4), dtype='float32')
a_f64 = np.zeros((3, 4), dtype='float64')
a_bool = np.zeros((3, 4), dtype='bool')

# ## Strings
good('array', a_f32)
good('array', a_f64)
# synonims
good('ndarray', a_f32, exact=False)  # will be canonicalized to "array"
good('ndarray', a_f64, exact=False)  # same
fail('array', [0, 1])
# dtypes
good('array(uint8)', a_u8)
good('array(u1)', a_u8)
good('array(int8)', a_i8)
good('array(i1)', a_i8)
good('array(float32)', a_f32)
good('array(float64)', a_f64)
good('array(bool)', a_bool)
fail('array(float64)', a_f32)
fail('array(float32)', a_f64)
fail('array(uint8)', a_f32)
fail('array(float32)', a_u8)

# shapes
a0d = np.zeros((), dtype='float32')
a1d = np.zeros((2,))
a2d = np.zeros((2, 4,))
a3d = np.zeros((2, 4, 8))

good('shape[0]', a0d)
good('shape[1]', a1d)
good('shape[2]', a2d)
good('shape[3]', a3d)
fail('shape[>0]', a0d)
fail('shape[<1]', a1d)
fail('shape[>2]', a2d)
fail('shape[<3]', a3d)
good(['shape[x]', 'shape[y],x=y'], [a3d, a3d])
good(['shape[x]', 'shape[y],x=y'], [a2d, a2d])
good(['shape[x]', 'shape[y],x=y'], [a1d, a1d])
good(['shape[x]', 'shape[y],x=y'], [a0d, a0d])

good('array[2x4]', a2d)
fail('array[AxBxC]', a2d)
fail('array[2x4]', a3d)

good('array[HxW],H=2,W>3', a2d)
good('array[(=2)x(>3)]', a2d, exact=False)  # Parenthesis are unnecessary

# ellipsis to mean 0 or more dimensions 
good('array[2x4x...]', a2d)
good('array[2x4x...]', a3d)

# if we really want more, use:
good('shape[>2],array[2x4x...]', a3d)
fail('shape[>2],array[2x4x...]', a2d)

fail('shape[>2]', [2, 2, 3])

# Try some binding:
good('array[XxYx...],X=2,Y=4', a3d)

# We don't do in between yet
syntax_fail('array[2x...x3]')

# We don't do in between yet
syntax_fail('array[2x...x3]')
# Try some binding:
good('array[XxYx...],X=2,Y=4', a2d)
good('array[XxYx...],X=2,Y=4', a3d)

# Using the array syntax  array[shape desc]

v1d = np.zeros(100)
v2d = np.zeros((10, 10))
good('array[100]', v1d)
fail('array[100]', v2d)
good('array[10x10]', v2d)
fail('array[10x...]', v1d)
good('array[10x...]', v2d)

fail('shape[>0]', a0d)
fail('shape[<1]', a1d)
fail('shape[>2]', a2d)
fail('shape[<3]', a3d)
good('shape[0]', a0d)
good('shape[1]', a1d)
good('shape[2]', a2d)
good('shape[3]', a3d)
good('array[shape[3]]', a3d)

good('array[1x2]', zeros((1, 2)))

# Now: special comparisons for arrays
A = np.array([0, 1, 2])
B = np.array([10, 20, 30])
good('array(>=0)', A)
good('array(<=2)', A)
fail('array(<=20)', B)
good('array(=0)', zeros((10)))
fail('array(=0)', np.array([0, 1, 0]))
good('array(=1)', ones((10)))

good(['shape(x)', 'shape(x)'], [a2d, a2d])
fail(['shape(x)', 'shape(x)'], [a2d, a3d])

good('array[NxN](<=1)', np.ones((10, 10)))
good('array[NxN](<=1,float32)', np.ones((10, 10), dtype='float32'))
good('array[NxN](<=1,float32|float64)', np.ones((10, 10), dtype='float64'))
good('array[NxN](<=1,(float32|float64))',
     np.ones((10, 10), dtype='float64'))
good('array[NxN](<=1,>=1)', np.ones((10, 10)))

# more complicated tests

good('array[(2|3)xN]', np.ones((2, 10)))
good('array[(2|3)xN]', np.ones((3, 10)))
fail('array[(2|3)xN]', np.ones((4, 10)))
good('array[(2|3)x...]', np.ones((2, 10)))
good('array[(2|3)x...]', np.ones((3, 10)))
fail('array[(2|3)x...]', np.ones((4, 10)))

good('array[(2,*)xN]', np.ones((2, 10)))
fail('array[(2,3)xN]', np.ones((4, 10)))
good('array[(2,*)x...]', np.ones((2, 10)))
fail('array[(2,3)x...]', np.ones((4, 10)))


good('seq', np.ones(3))
good('seq[3]', np.ones(3))
fail('seq[3]', np.ones(2))
good('seq[6]', np.ones((2, 3)))
fail('seq[6]', np.ones((2, 4)))

# "finite" contract 
good('finite', 1)
good('finite', 0)
good('finite', -1)
good('finite', np.float_(1))
fail('finite', np.inf)
fail('finite', np.nan)

# some more complex stuff
good('array[N](>=-pi,<pi)', np.array([0], 'float32'))
good('array[N](>=0)', np.array([0], 'float32'))

# generalized ideas of numbers
# np_ints = ['int8', 'int16', 'int32', 'int64',
#            'uint8', 'uint16', 'uint32', 'uint64',
#            ]
# np_floats = ['float32', 'float64']

good('array(float)', np.array(1.32, 'float32'))


for dt in np_int_dtypes:
    x = np.array(1).astype(dt)
    good('number', x)
    fail('Number', x)

    good('array(int)', x)
    good('int', x)
    fail('Int', x)

    fail('array(float)', x)
    fail('float', x)
    fail('Float', x)
    
# generalized ideas of numbers
for dt in np_float_dtypes:
    x = np.array(1).astype(dt)
    good('number', x)
    fail('Number', x)
    
    fail('array(int)', x)
    fail('int', x)
    fail('Int', x)

    good('array(float)', x)
    good('float', x)
    fail('Float', x)
    
# generalized ideas of numbers
for dt in np_uint_dtypes:
    x = np.array(1).astype(dt)
    good('number', x)
    fail('Number', x)
    
    fail('array(int)', x)
    fail('int', x)
    fail('Int', x)

    good('array(uint)', x)
    fail('Int', x)

    fail('array(float)', x)
    fail('float', x)
    fail('Float', x)




"""Disabled as of 2.1.6 but it does seem to work.Except with CD
"""
# TODO Fix dearpygui module not found
# import dearpygui.dearpygui as dpg
#
# import sys
# this = sys.modules[__name__]
#
# for i in dir(dpg):
#     setattr(this, i, getattr(dpg, i))



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from abc import abstractmethod

import numpy as np

from .types_misc import CheckType
from ..interface import Contract, ContractNotRespected, RValue, eval_in_context
from ..syntax import W, Keyword, add_contract, add_keyword


class ArrayElementsTest(Contract):

    @abstractmethod
    def test_elements(self, context, value):
        """ Returns either a bool or an array of bool. """

    def check_contract(self, context, value, silent):
        result = self.test_elements(context, value)

        if np.all(result):
            return

        result = np.array(result)  # for simple bool
        resultf = result.flatten()
        valuef = value.flatten()
        some, = np.nonzero(np.logical_not(resultf))
        num = value.size
        num_fail = len(some)
        perc = 100.0 * num_fail / num
        error = ("In this array, %d/%d (%f%%) of elements do not respect "
                 "the condition %s." % (num_fail, num, perc, self))
        some_failures = valuef[some]
        MAX_N = 4
        if len(some_failures) > MAX_N:
            some_failures = some_failures[:MAX_N]
        failures = list(some_failures)
        N = len(failures)
        error += '\nThese are the first %d: %s.' % (N, failures)
        raise ContractNotRespected(self, error, value, context)


class ArrayLogical(ArrayElementsTest):
    def __init__(self, glyph, precedence):
        self.glyph = glyph
        self.precedence = precedence

    def __str__(self):
        def convert(x):
            if isinstance(x, ArrayLogical) and x.precedence < self.precedence:
                return '(%s)' % x
            else:
                return '%s' % x

        s = self.glyph.join(convert(x) for x in self.clauses)
        return s


class ArrayOR(ArrayLogical):
    def __init__(self, clauses, where=None):
        assert isinstance(clauses, list)
        assert len(clauses) >= 2
        for c in clauses:
            assert isinstance(c, ArrayElementsTest)
        Contract.__init__(self, where)
        ArrayLogical.__init__(self, '|', 1)
        self.clauses = clauses

    def test_elements(self, context, value):
        assert isinstance(value, np.ndarray)
        result = False
        for c in self.clauses:
            result_c = c.test_elements(context, value)
            result = np.logical_or(result_c, result)
        return result

    def __repr__(self):
        s = 'ArrayOR(%r)' % self.clauses
        return s

    @staticmethod
    def parse_action(string, location, tokens):
        l = list(tokens[0])
        clauses = [l.pop(0)]
        while l:
            glyph = l.pop(0)  # @UnusedVariable
            assert glyph == '|'
            operand = l.pop(0)
            clauses.append(operand)
        where = W(string, location)
        return ArrayOR(clauses, where=where)


class ArrayORCustomString(ArrayOR):
    def __init__(self, custom_string, **other):
        self.custom_string = custom_string
        ArrayOR.__init__(self, **other)

    def __str__(self):
        return self.custom_string


class ArrayAnd(ArrayLogical):
    def __init__(self, clauses, where=None):
        assert isinstance(clauses, list)
        assert len(clauses) >= 2, clauses
        assert isinstance(clauses, list)
        assert len(clauses) >= 2
        for c in clauses:
            assert isinstance(c, ArrayElementsTest)

        Contract.__init__(self, where)
        ArrayLogical.__init__(self, ',', 2)
        self.clauses = clauses

    def test_elements(self, context, value):
        assert isinstance(value, np.ndarray)
        result = True
        for c in self.clauses:
            result_c = c.test_elements(context, value)
            result = np.logical_and(result_c, result)
        return result

    def __repr__(self):
        s = 'ArrayAnd(%r)' % self.clauses
        return s

    @staticmethod
    def parse_action(string, location, tokens):
        l = list(tokens[0])
        clauses = [l.pop(0)]
        while l:
            glyph = l.pop(0)  # @UnusedVariable
            assert glyph == ','
            operand = l.pop(0)
            clauses.append(operand)
        where = W(string, location)
        return ArrayAnd(clauses, where=where)


class ArrayConstraint(ArrayElementsTest):
    """ Comparisons for numpy array elements. They check that
        the condition is respected for all the entries in the array. """

    constraints = {
        '=': lambda x, rvalue: x == rvalue,
        '==': lambda x, rvalue: x == rvalue,
        '!=': lambda x, rvalue: x != rvalue,
        '>': lambda x, rvalue: x > rvalue,
        '>=': lambda x, rvalue: x >= rvalue,
        '<': lambda x, rvalue: x < rvalue,
        '<=': lambda x, rvalue: x <= rvalue,
    }

    def __init__(self, glyph, rvalue, where=None):
        assert glyph in ArrayConstraint.constraints
        assert isinstance(rvalue, RValue)
        Contract.__init__(self, where)
        self.glyph = glyph
        self.rvalue = rvalue

    def test_elements(self, context, value):
        """ Returns either a bool or an array of bool. """
        assert isinstance(value, np.ndarray)
        bound = eval_in_context(context=context, value=self.rvalue,
                                contract=self)

        operation = ArrayConstraint.constraints[self.glyph]
        result = operation(value, bound)
        return result

    def __str__(self):
        return '%s%s' % (self.glyph, self.rvalue)

    def __repr__(self):
        return 'ArrayConstraint(%r,%r)' % (self.glyph, self.rvalue)

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        glyph = "".join(tokens['glyph'])
        rvalue = tokens['rvalue']
        return ArrayConstraint(glyph, rvalue, where)


class DType(ArrayElementsTest):
    """ Checks that the value is an array with the given dtype. """

    def __init__(self, dtype, dtype_string=None, where=None):
        assert isinstance(dtype, np.dtype)
        Contract.__init__(self, where)
        self.dtype = dtype
        if dtype_string is None:
            dtype_string = "%s" % dtype
        self.dtype_string = dtype_string

    def test_elements(self, context, value):  # @UnusedVariable
        assert isinstance(value, np.ndarray)  # Guaranteed by construction
        return value.dtype == self.dtype

    def __str__(self):
        return self.dtype_string

    def __repr__(self):
        if "%s" % self.dtype == self.dtype_string:
            return 'DType(%r)' % self.dtype
        else:
            return 'DType(%r,%r)' % (self.dtype, self.dtype_string)

    @staticmethod
    def parse_action(dtype=None):
        assert dtype is None or isinstance(dtype, np.dtype)

        def parse(s, loc, tokens):
            where = W(s, loc)
            dtype_string = tokens[0]
            if dtype is None:
                use_dtype = np.dtype(dtype_string)
            else:
                use_dtype = dtype
            return DType(use_dtype, dtype_string, where)

        return parse


np_types = {
    'np_int': np.int_,  # Platform integer (normally either int32 or int64)
    'np_int8': np.int8,  # Byte (-128 to 127)
    'np_int16': np.int16,  # Integer (-32768 to 32767)
    'np_int32': np.int32,  # Integer (-2147483648 to 2147483647)
    'np_int64': np.int64,  # Integer (9223372036854775808 to 9223372036854775807)
    'np_uint8': np.uint8,  # Unsigned integer (0 to 255)
    'np_uint16': np.uint16,  # Unsigned integer (0 to 65535)
    'np_uint32': np.uint32,  # Unsigned integer (0 to 4294967295)
    'np_uint64': np.uint64,  # Unsigned integer (0 to 18446744073709551615)
    'np_float': np.float_,  # Shorthand for float64.
    'np_float16': np.float16,  # Half precision float: sign bit, 5 bits exponent, 10 bits mantissa
    'np_float32': np.float32,  # Single precision float: sign bit, 8 bits exponent, 23 bits mantissa
    'np_float64': np.float64,  # Double precision float: sign bit, 11 bits exponent, 52 bits mantissa
    'np_complex': np.complex_,  # Shorthand for complex128.
    'np_complex64': np.complex64,  # Complex number, represented by two 32-bit floats (real and imaginary components)
    'np_complex128': np.complex128}

for k, t in np_types.items():
    add_contract(Keyword(k).setParseAction(CheckType.parse_action(t)))
    add_keyword(k)



__version__ = '2.0.1'

from .interface import (Contract, ContractNotRespected,
                        CannotDecorateClassmethods,
                        ContractSyntaxError, ContractException)

from .main import (contract_decorator, check, fail, check_multiple,
                   contracts_decorate as decorate,
                   parse_flexible_spec as parse)


# Just make them appear as belonging to the "contracts" Module
# So that Eclipse and other IDEs will not get confused.
contract = contract_decorator  # type: ignore

contract.__doc__ = contract_decorator.__doc__

from .main import new_contract as new_contract_main


def new_contract(*args):
    return new_contract_main(*args)


new_contract.__doc__ = new_contract_main.__doc__

# A couple of useful functions
from .interface import describe_value, describe_type, describe_value_multiline
from .utils import *

from .metaclass import ContractsMeta, with_metaclass

ContractsMeta.__module__ = 'contracts'

# And after everything else is loaded, load the  utils
from .useful_contracts import *
# After everything is loaded, load aliases
from .library import miscellaneous_aliases  # @UnusedImport



import decimal
import operator as ops
import re
from typing import Callable, Optional, Dict, Any

from ._util import get_arg_types, get_return_type, get_kwarg_types
from .core import Pattern, MatchContext, MatchResult, StringPattern, OneOf, Nested, Underscore


class Check(Pattern):
    def __init__(self, condition):
        self._condition = condition

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        return ctx.match_if(self._condition(value))


class Regex(Pattern, StringPattern):
    def __init__(self, regex, *, bind_groups: bool = True, capture_wildcards: bool = False):
        self._regex: re.Pattern = re.compile(regex)
        self._bind_groups = bind_groups
        if capture_wildcards:
            self._wildcards = tuple(Underscore() for _i in range(0, self._regex.groups))
        else:
            self._wildcards = tuple([])

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        try:
            result = self._regex.fullmatch(value)
        except TypeError:
            return ctx.no_match()
        if not result:
            return ctx.no_match()
        if self._bind_groups:
            for k, v in result.groupdict().items():
                ctx[k] = v
        for ix, wc in enumerate(self._wildcards):
            ctx.record(wc, result.group(ix + 1))
        return ctx.matches()

    def string_match(self, remaining, *, ctx: MatchContext) -> Optional[str]:
        result = self._regex.match(remaining)
        if not result:
            return None
        if self._bind_groups:
            for k, v in result.groupdict().items():
                ctx[k] = v
        return result.group(0)


class InstanceOf(Pattern):
    def __init__(self, *type_: type):
        self._type = type_

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        return ctx.match_if(isinstance(value, self._type))


class SubclassOf(Pattern):
    def __init__(self, *type_: type):
        self._type = type_

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        return ctx.match_if(issubclass(value, self._type))


class Between(Pattern):
    def __init__(self, lower, upper, *, lower_bound_exclusive=False, upper_bound_exclusive=False):
        self.lower = lower
        self.upper = upper
        self.op_lower = ops.gt if lower_bound_exclusive else ops.ge
        self.op_upper = ops.lt if upper_bound_exclusive else ops.le

    def match(self, value, *, ctx: MatchContext, strict=False) -> MatchResult:
        return ctx.match_if(self.op_lower(value, self.lower) and self.op_upper(value, self.upper))


class Length(Pattern):
    def __init__(self, *, exactly: int = None, at_least: int = None, at_most: int = None):
        if exactly is not None:
            if at_least is not None or at_most is not None:
                raise ValueError("If length is given, 'at_least' or 'at_most' must not be given.")
            self._at_least = exactly
            self._at_most = exactly
        else:
            self._at_least = at_least
            self._at_most = at_most

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        length = len(value)
        return ctx.match_if((self._at_least is None or length >= self._at_least)
                            and (self._at_most is None or length <= self._at_most))


class Contains(Pattern):
    def __init__(self, needle):
        self._needle = needle

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        return ctx.match_if(self._needle in value)


class Transformed(Pattern, Nested):
    def __init__(self, f: Callable, pattern):
        self._f = f
        self._pattern = pattern

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        # noinspection PyBroadException
        try:
            transformed = self._f(value)
        except Exception:
            return ctx.no_match()
        return ctx.match(transformed, self._pattern)

    def descend(self, f):
        return Transformed(f=self._f, pattern=f(self._pattern))


class Arguments(Pattern, Nested):
    def __init__(self, *arg_patterns, **kwargs):
        self._pattern = list(arg_patterns)
        self._kwargs: Dict[str, Any] = kwargs

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        if not callable(value):
            return ctx.no_match()
        if self._pattern:
            result = ctx.match(get_arg_types(value), self._pattern)
            if not result:
                return result
        return ctx.match(get_kwarg_types(value), self._kwargs, strict)

    def descend(self, f):
        t_pattern = list(f(p) for p in self._pattern)
        t_kwargs = {}
        for k, v in self._kwargs.items():
            t_kwargs[k] = f(v)
        return Arguments(*t_pattern, **t_kwargs)


class Returns(Pattern, Nested):
    def __init__(self, pattern):
        self._pattern = pattern

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        if not callable(value):
            return ctx.no_match()
        return ctx.match(get_return_type(value), self._pattern)

    def descend(self, f):
        return Returns(pattern=f(self._pattern))


class Each(Pattern, Nested):
    def __init__(self, pattern, *, at_least: int = 0):
        self._pattern = pattern
        self._at_least = at_least

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        count = 0
        try:
            it = iter(value)
        except TypeError:
            return ctx.no_match()
        for item in it:
            result = ctx.match(item, self._pattern)
            if not result:
                return result
            count += 1
        return ctx.match_if(count >= self._at_least)

    def descend(self, f):
        return Each(pattern=f(self._pattern), at_least=self._at_least)


class EachItem(Pattern, Nested):
    def __init__(self, key_pattern, value_pattern):
        self._key_pattern = key_pattern
        self._value_pattern = value_pattern

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        try:
            items = value.items()
        except AttributeError:
            return ctx.no_match()
        for k, v in items:
            result = ctx.match(k, self._key_pattern)
            if not result:
                return result
            result = ctx.match(v, self._value_pattern)
            if not result:
                return result
        return ctx.matches()

    def descend(self, f):
        return EachItem(key_pattern=f(self._key_pattern), value_pattern=f(self._value_pattern))


class At(Pattern, Nested):
    def __init__(self, path, pattern):
        if isinstance(path, str):
            self._path = path.split(".")
        else:
            self._path = list(path)
        self._pattern = pattern

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        for k in self._path:
            try:
                value = value[k]
            except (TypeError, KeyError):
                return ctx.no_match()
        return ctx.match(value, self._pattern)

    def descend(self, f):
        return At(path=self._path, pattern=f(self._pattern))


class Items(Pattern, Nested):
    def __init__(self, **kwargs):
        self._items = kwargs

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        return ctx.match(value, self._items, strict=strict)

    def descend(self, f):
        items = {}
        for k, v in self._items.items():
            items[k] = f(v)
        return Items(**items)


class Attrs(Pattern, Nested):
    def __init__(self, **kwargs):
        self._items = kwargs

    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
        try:
            attrs = value.__dict__
        except AttributeError:
            return ctx.no_match()
        return ctx.match(attrs, self._items, strict=strict)

    def descend(self, f):
        items = {}
        for k, v in self._items.items():
            items[k] = f(v)
        return Attrs(**items)


# noinspection PyPep8Naming
def NoneOf(*args) -> Pattern:
    return ~OneOf(args)


# noinspection PyPep8Naming
def Maybe(pattern) -> Pattern:
    return OneOf(pattern, ...)


IsTruish = Check(lambda x: bool(x))
IsNumber = (InstanceOf(int) & ~InstanceOf(bool)) | InstanceOf(float) | InstanceOf(decimal.Decimal)
IsString = InstanceOf(str)



import numbers
import unittest

from Aspidites._vendor.apm import *


class ReportingTest(unittest.TestCase):
    instance_of = InstanceOf(numbers.Number)
    each = Each(instance_of)

    def test_match_stack(self):
        result = match([1, 2, "uvw", 4], self.each)
        self.assertFalse(result)
        self.assertEqual(([1, 2, 'uvw', 4], 'uvw'), tuple(k for k, _v in result._match_stack))
        self.assertEqual((
            self.each,
            self.instance_of,
        ), tuple(v for _k, v in result._match_stack))

    def test_explain(self):
        result = match([1, 2, "uvw", 4], self.each)
        self.assertEqual(
            f"[1, 2, 'uvw', 4]\n"
            f"...did not match the pattern:\n"
            f"{repr(self.each)}\n"
            f"...because:\n"
            f"uvw\n"
            f"...did not match the pattern:\n"
            f"{repr(self.instance_of)}", result.explain())

    def test_explain_short(self):
        result = match([1, 2, "uvw", 4], self.each)
        self.assertEqual(
            f"uvw\n"
            f"...did not match the pattern:\n"
            f"{repr(self.instance_of)}", result.explain(short=True))

    def test_explain_matching(self):
        result = match(1, self.instance_of)
        self.assertTrue(result)
        self.assertEqual("The pattern matches the given value.", result.explain())


if __name__ == '__main__':
    unittest.main()



#
#  nested.py
#  Copyright, 2007 - Paul McGuire
#
#  Simple example of using nestedExpr to define expressions using
#  paired delimiters for grouping lists and sublists
#

from pyparsing import *

data = """
{
     { item1 "item with } in it" }
     {
      {item2a item2b }
      {item3}
     }

}
"""

# use {}'s for nested lists
nestedItems = nestedExpr("{", "}")
print(( (nestedItems+stringEnd).parseString(data).asList() ))

# use default delimiters of ()'s
mathExpr = nestedExpr()
print(( mathExpr.parseString( "((( ax + by)*C) *(Z | (E^F) & D))") ))



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected
from ..syntax import W, contract_expression, add_contract, add_keyword, Keyword
from Aspidites._vendor.pyparsing import (Dict, delimitedList, Group, alphanums, Suppress, Literal,
    Word)


class Attr(Contract):
    def __init__(self, attrs, where=None):
        # attrs: dict(str: contract)
        Contract.__init__(self, where)
        if not attrs:
            raise ValueError('Need some attributes') 
        self.attrs = attrs

    def check_contract(self, context, value, silent):
        for k in self.attrs:
            if hasattr(value, k):
                self.attrs[k]._check_contract(context, getattr(value, k), silent)
            else:
                error = 'Expected an attribute %r.' % k
                raise ContractNotRespected(contract=self, error=error,
                                           value=value, context=context)


    def __str__(self):
        sattrs = ";".join(['%s:%s' % (k, self.attrs[k]) for k in sorted(self.attrs)]) 
        return 'attr(%s)' % sattrs

    def __repr__(self):
        rattrs = ",".join(['%r:%r' % (k, self.attrs[k]) for k in sorted(self.attrs)]) 
        return 'Attr({%s})' % rattrs

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        attrs = tokens.get('attrs', None)
        # Python 3 only
        # attrs = {k: v for k, v in attrs.items()}
        attrs = dict([(k, v) for k, v in attrs.items()])
        return Attr(attrs, where=where)

attr_spec = Dict(
        delimitedList(
            Group(Word(alphanums) + 
                    Suppress(Literal(':')) + 
                    contract_expression('value')),
            delim=';')
        )('attrs')
attrs_spec = ('(' - attr_spec - ')')


attr_contract = Keyword('attr') - attrs_spec
attr_contract.setParseAction(Attr.parse_action)

add_contract(attr_contract)
add_keyword('attr')



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval

def test_In():
    assert restricted_eval('1 in [1, 2, 3]') is True


def test_NotIn():
    assert restricted_eval('4 not in [1, 2, 3]') is True



# parseActions.py
#
#   A sample program a parser to match a date string of the form "YYYY/MM/DD",
# and return it as a datetime, or raise an exception if not a valid date.
#
# Copyright 2012, Paul T. McGuire
#
from datetime import datetime
from ....._vendor import pyparsing as pp
from ....._vendor.pyparsing import pyparsing_common as ppc

# define an integer string, and a parse action to convert it
# to an integer at parse time
integer = pp.Word(pp.nums).setName("integer")
def convertToInt(tokens):
    # no need to test for validity - we can't get here
    # unless tokens[0] contains all numeric digits
    return int(tokens[0])
integer.setParseAction(convertToInt)
# or can be written as one line as
#integer = Word(nums).setParseAction(lambda t: int(t[0]))

# define a pattern for a year/month/day date
date_expr = integer('year') + '/' + integer('month') + '/' + integer('day')
date_expr.ignore(pp.pythonStyleComment)

def convertToDatetime(s,loc,tokens):
    try:
        # note that the year, month, and day fields were already
        # converted to ints from strings by the parse action defined
        # on the integer expression above
        return datetime(tokens.year, tokens.month, tokens.day).date()
    except Exception as ve:
        errmsg = "'%s/%s/%s' is not a valid date, %s" % \
            (tokens.year, tokens.month, tokens.day, ve)
        raise pp.ParseException(s, loc, errmsg)
date_expr.setParseAction(convertToDatetime)


date_expr.runTests("""\
    2000/1/1

    # invalid month
    2000/13/1

    # 1900 was not a leap year
    1900/2/29

    # but 2000 was
    2000/2/29
    """)


# if dates conform to ISO8601, use definitions in pyparsing_common
date_expr = ppc.iso8601_date.setParseAction(ppc.convertToDate())
date_expr.ignore(pp.pythonStyleComment)

date_expr.runTests("""\
    2000-01-01

    # invalid month
    2000-13-01

    # 1900 was not a leap year
    1900-02-29

    # but 2000 was
    2000-02-29
    """)



from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor._compat import IS_PY2
from Aspidites._vendor._compat import IS_PY3
from Aspidites._vendor.RestrictedPython.Eval import default_guarded_getiter
from Aspidites._vendor.RestrictedPython.Guards import guarded_unpack_sequence
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_exec

import pytest


lambda_err_msg = 'Line 1: "_bad" is an invalid variable ' \
                 'name because it starts with "_"'


def test_RestrictingNodeTransformer__visit_Lambda__1():
    """It prevents arguments starting with `_`."""
    result = compile_restricted_exec("lambda _bad: None")
    # Aspidites.libraries.RestrictedPython.compile.compile_restricted_exec on Python 2 renders
    # the error message twice. This is necessary as otherwise *_bad and **_bad
    # would be allowed.
    assert lambda_err_msg in result.errors


def test_RestrictingNodeTransformer__visit_Lambda__2():
    """It prevents keyword arguments starting with `_`."""
    result = compile_restricted_exec("lambda _bad=1: None")
    # Aspidites.libraries.RestrictedPython.compile.compile_restricted_exec on Python 2 renders
    # the error message twice. This is necessary as otherwise *_bad and **_bad
    # would be allowed.
    assert lambda_err_msg in result.errors


def test_RestrictingNodeTransformer__visit_Lambda__3():
    """It prevents * arguments starting with `_`."""
    result = compile_restricted_exec("lambda *_bad: None")
    assert result.errors == (lambda_err_msg,)


def test_RestrictingNodeTransformer__visit_Lambda__4():
    """It prevents ** arguments starting with `_`."""
    result = compile_restricted_exec("lambda **_bad: None")
    assert result.errors == (lambda_err_msg,)


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple parameter unpacking is gone in Python 3")
def test_RestrictingNodeTransformer__visit_Lambda__5():  # pragma: no cover
    """It prevents arguments starting with `_` in tuple unpacking."""
    result = compile_restricted_exec("lambda (a, _bad): None")
    # Aspidites.libraries.RestrictedPython.compile.compile_restricted_exec on Python 2 renders
    # the error message twice. This is necessary as otherwise *_bad and
    # **_bad would be allowed.
    assert lambda_err_msg in result.errors


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple parameter unpacking is gone in Python 3")
def test_RestrictingNodeTransformer__visit_Lambda__6():  # pragma: no cover
    """It prevents arguments starting with `_` in nested tuple unpacking."""
    result = compile_restricted_exec("lambda (a, (c, (_bad, c))): None")
    # Aspidites.libraries.RestrictedPython.compile.compile_restricted_exec on Python 2 renders
    # the error message twice. This is necessary as otherwise *_bad and
    # **_bad would be allowed.
    assert lambda_err_msg in result.errors


@pytest.mark.skipif(
    IS_PY2,
    reason="There is no single `*` argument in Python 2")
def test_RestrictingNodeTransformer__visit_Lambda__7():  # pragma: PY3
    """It prevents arguments starting with `_` together with a single `*`."""
    result = compile_restricted_exec("lambda good, *, _bad: None")
    assert result.errors == (lambda_err_msg,)


BAD_ARG_IN_LAMBDA = """\
def check_getattr_in_lambda(arg=lambda _bad=(lambda ob, name: name): _bad2):
    42
"""


def test_RestrictingNodeTransformer__visit_Lambda__8():
    """It prevents arguments starting with `_` in weird lambdas."""
    result = compile_restricted_exec(BAD_ARG_IN_LAMBDA)
    # On Python 2 the first error message is contained twice:
    assert lambda_err_msg in result.errors


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple parameter unpacking is gone in python 3")
def test_RestrictingNodeTransformer__visit_Lambda__9(
        mocker):  # pragma: no cover
    _getiter_ = mocker.stub()
    _getiter_.side_effect = lambda it: it

    glb = {
        '_getiter_': _getiter_,
        '_unpack_sequence_': guarded_unpack_sequence,
        '_getattr_': getattr,
        'sum': sum,
    }

    src = "m = lambda (a, (b, c)), *ag, **kw: a+b+c+sum(ag)+sum(kw.values())"
    restricted_exec(src, glb)

    ret = glb['m']((1, (2, 3)), 4, 5, 6, g=7, e=8)
    assert ret == 36
    assert 2 == _getiter_.call_count
    _getiter_.assert_any_call((1, (2, 3)))
    _getiter_.assert_any_call((2, 3))


LAMBDA_FUNC_1 = """
g = lambda x: x ** 2
"""


def test_RestrictingNodeTransformer__visit_Lambda__10():
    """Simple lambda functions are allowed."""
    restricted_globals = dict(
        g=None,
    )
    restricted_exec(LAMBDA_FUNC_1, restricted_globals)
    assert 4 == restricted_globals['g'](2)


LAMBDA_FUNC_2 = """
g = lambda (x, y) : (x ** 2, x + y)
"""


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple parameter unpacking is gone in python 3")
def test_RestrictingNodeTransformer__visit_Lambda__11():  # pragma: no cover
    """Lambda functions with tuple unpacking are allowed."""
    restricted_globals = dict(
        g=None,
        _unpack_sequence_=guarded_unpack_sequence,
        _getiter_=default_guarded_getiter,
    )
    restricted_exec(LAMBDA_FUNC_2, restricted_globals)
    assert (9, 5) == restricted_globals['g']((3, 2))


LAMBDA_FUNC_3 = """
g = lambda (x, y), z : (x ** y, x + z)
"""


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple parameter unpacking is gone in python 3")
def test_RestrictingNodeTransformer__visit_Lambda__12():  # pragma: no cover
    """Lambda functions with tuple unpacking and simple params are allowed."""
    restricted_globals = dict(
        g=None,
        _unpack_sequence_=guarded_unpack_sequence,
        _getiter_=default_guarded_getiter,
    )
    restricted_exec(LAMBDA_FUNC_3, restricted_globals)
    assert (64, 6) == restricted_globals['g']((4, 3), 2)



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
import operator


def find_longest_match(s, options):
    matches = [(x, longest_match(s, x)) for x in options]
    best = max(matches, key=operator.itemgetter(1))
    return best


def longest_match(a, b):
    lengths = range(min(len(a), len(b)) + 1)
    lengths = list(reversed(lengths))
    for i in lengths:
        if a[:i] == b[:i]:
            return i
    assert False  # pragma: no cover


assert ('float64', 6) == find_longest_match('float6', ['float32', 'float64'])
assert 2 == find_longest_match('fl6', ['float32', 'float64'])[1]


# http://hetland.org/coding/python/levenshtein.py
def levenshtein(a, b):
    """Calculates the Levenshtein distance between a and b."""
    n, m = len(a), len(b)
    if n > m:  # pragma: no cover
        # Make sure n <= m, to use O(min(n,m)) space
        a, b = b, a
        n, m = m, n

    current = range(n + 1)
    for i in range(1, m + 1):
        previous, current = current, [i] + [0] * n
        for j in range(1, n + 1):
            add, delete = previous[j] + 1, current[j - 1] + 1
            change = previous[j - 1]
            if a[j - 1] != b[i - 1]:
                change = change + 1
            current[j] = min(add, delete, change)

    return current[n]


def find_best_match(s, options):
    matches = [(x, levenshtein(s[:len(x)], x) - len(x)) for x in options]
    best = min(matches, key=operator.itemgetter(1))
    return best


def default_message(identifier):
    return 'Unknown identifier %r. ' % identifier


def create_suggester(get_options, get_message=default_message,
            pattern=None):

    from ..syntax import (Combine, Word, alphas, alphanums, oneOf,
                      ParseSyntaxException, ParseException)
    if pattern is None:
        pattern = Combine(oneOf(list(alphas)) + Word('_' + alphanums))


    pattern = pattern.copy()

    def find_match(identifier, options, local_string):
        match, length = find_longest_match(identifier, options)
        if length >= 2:
            return True, match, length

        match, distance = find_best_match(local_string, options)

        if distance < len(match) - 1:
            length = longest_match(local_string, match)
            return True, match, length

        return False, "No matches found", 0

    def parse_action(s, loc, tokens):
        identifier = tokens[0]
        options = get_options()
        
        msg = 'Bug in syntax: I was not supposed to match %r.' % identifier
        msg += '(options: %s)' % options
        
        msg += ''' Suggestions on the cause:
            1) Use add_keyword(), always.
            
            2) Use:
                Keyword('attr') - attrs_spec
               instead of 
                Keyword('attr') + attrs_spec
        '''
        assert not (identifier in options), msg
        msg = get_message(identifier)

        if options:
            local_string = s[loc:]
            confident, match, length = find_match(identifier, options,
                                                  local_string)
            if confident:
                msg += 'Did you mean %r?' % match
                loc += length
            else:
                msg += '\nI know: %r.\n' % (options)
        pe = ParseException(s, loc, msg, pattern)
        raise ParseSyntaxException(pe)

    pattern.setParseAction(parse_action)

    return pattern



"""
This module can parse a Delphi Form (dfm) file.

The main is used in experimenting (to find which files fail
to parse, and where), but isn't useful for anything else.
"""
__version__ = "1.0"
__author__ = "Daniel 'Dang' Griffith <pythondev - dang at lazytwinacres . net>"


from ....._vendor.pyparsing import Literal, CaselessLiteral, Word, delimitedList \
    , Optional, Combine, Group, alphas, nums, alphanums, Forward \
    , oneOf, OneOrMore, ZeroOrMore, CharsNotIn


# This converts DFM character constants into Python string (unicode) values.
def to_chr(x):
    """chr(x) if 0 < x < 128 ; unicode(x) if x > 127."""
    return 0 < x < 128 and chr(x) or eval("u'\\u%d'" % x )

#################
# BEGIN GRAMMAR
#################

COLON  = Literal(":").suppress()
CONCAT = Literal("+").suppress()
EQUALS = Literal("=").suppress()
LANGLE = Literal("<").suppress()
LBRACE = Literal("[").suppress()
LPAREN = Literal("(").suppress()
PERIOD = Literal(".").suppress()
RANGLE = Literal(">").suppress()
RBRACE = Literal("]").suppress()
RPAREN = Literal(")").suppress()

CATEGORIES  = CaselessLiteral("categories").suppress()
END         = CaselessLiteral("end").suppress()
FONT        = CaselessLiteral("font").suppress()
HINT        = CaselessLiteral("hint").suppress()
ITEM        = CaselessLiteral("item").suppress()
OBJECT      = CaselessLiteral("object").suppress()

attribute_value_pair = Forward() # this is recursed in item_list_entry

simple_identifier = Word(alphas, alphanums + "_")
identifier = Combine( simple_identifier + ZeroOrMore( Literal(".") + simple_identifier ))
object_name = identifier
object_type = identifier

# Integer and floating point values are converted to Python longs and floats, respectively.
int_value = Combine(Optional("-") + Word(nums)).setParseAction(lambda s,l,t: [ int(t[0]) ] )
float_value = Combine(Optional("-") + Optional(Word(nums)) + "." + Word(nums)).setParseAction(lambda s,l,t: [ float(t[0]) ] )
number_value = float_value | int_value

# Base16 constants are left in string form, including the surrounding braces.
base16_value = Combine(Literal("{") + OneOrMore(Word("0123456789ABCDEFabcdef")) + Literal("}"), adjacent=False)

# This is the first part of a hack to convert the various delphi partial sglQuotedStrings
#     into a single sglQuotedString equivalent.  The gist of it is to combine
#     all sglQuotedStrings (with their surrounding quotes removed (suppressed))
#     with sequences of #xyz character constants, with "strings" concatenated
#     with a '+' sign.
unquoted_sglQuotedString = Combine( Literal("'").suppress() + ZeroOrMore( CharsNotIn("'\n\r") ) + Literal("'").suppress() )

# The parse action on this production converts repetitions of constants into a single string.
pound_char = Combine(
    OneOrMore((Literal("#").suppress()+Word(nums)
    ).setParseAction( lambda s, l, t: to_chr(int(t[0]) ))))

# This is the second part of the hack.  It combines the various "unquoted"
#     partial strings into a single one.  Then, the parse action puts
#     a single matched pair of quotes around it.
delphi_string = Combine(
    OneOrMore(CONCAT | pound_char | unquoted_sglQuotedString)
    , adjacent=False
    ).setParseAction(lambda s, l, t: "'%s'" % t[0])

string_value = delphi_string | base16_value

list_value = LBRACE + Optional(Group(delimitedList(identifier | number_value | string_value))) + RBRACE
paren_list_value = LPAREN + ZeroOrMore(identifier | number_value | string_value) + RPAREN

item_list_entry = ITEM + ZeroOrMore(attribute_value_pair) + END
item_list = LANGLE + ZeroOrMore(item_list_entry) + RANGLE

generic_value = identifier
value = item_list | number_value | string_value | list_value | paren_list_value | generic_value

category_attribute = CATEGORIES + PERIOD + oneOf("strings itemsvisibles visibles", True)
event_attribute = oneOf("onactivate onclosequery onclose oncreate ondeactivate onhide onshow", True)
font_attribute = FONT + PERIOD + oneOf("charset color height name style", True)
hint_attribute = HINT
layout_attribute = oneOf("left top width height", True)
generic_attribute = identifier
attribute = (category_attribute | event_attribute | font_attribute | hint_attribute | layout_attribute | generic_attribute)

category_attribute_value_pair = category_attribute + EQUALS + paren_list_value
event_attribute_value_pair = event_attribute + EQUALS + value
font_attribute_value_pair = font_attribute + EQUALS + value
hint_attribute_value_pair = hint_attribute + EQUALS + value
layout_attribute_value_pair = layout_attribute + EQUALS + value
generic_attribute_value_pair = attribute + EQUALS + value
attribute_value_pair << Group(
      category_attribute_value_pair
    | event_attribute_value_pair
    | font_attribute_value_pair
    | hint_attribute_value_pair
    | layout_attribute_value_pair
    | generic_attribute_value_pair
    )

object_declaration = Group((OBJECT + object_name + COLON + object_type))
object_attributes = Group(ZeroOrMore(attribute_value_pair))

nested_object = Forward()
object_definition = object_declaration + object_attributes + ZeroOrMore(nested_object) + END
nested_object << Group(object_definition)

#################
# END GRAMMAR
#################

def printer(s, loc, tok):
    print(tok, end=' ')
    return tok

def get_filename_list(tf):
    import sys, glob
    if tf == None:
        if len(sys.argv) > 1:
            tf = sys.argv[1:]
        else:
            tf = glob.glob("*.dfm")
    elif type(tf) == str:
        tf = [tf]
    testfiles = []
    for arg in tf:
        testfiles.extend(glob.glob(arg))
    return testfiles

def main(testfiles=None, action=printer):
    """testfiles can be None, in which case the command line arguments are used as filenames.
    testfiles can be a string, in which case that file is parsed.
    testfiles can be a list.
    In all cases, the filenames will be globbed.
    If more than one file is parsed successfully, a dictionary of ParseResults is returned.
    Otherwise, a simple ParseResults is returned.
    """
    testfiles = get_filename_list(testfiles)
    print(testfiles)

    if action:
        for i in (simple_identifier, value, item_list):
            i.setParseAction(action)

    success = 0
    failures = []

    retval = {}
    for f in testfiles:
        try:
            retval[f] = object_definition.parseFile(f)
            success += 1
        except Exception:
            failures.append(f)

    if failures:
        print('\nfailed while processing %s' % ', '.join(failures))
    print('\nsucceeded on %d of %d files' %(success, len(testfiles)))

    if len(retval) == 1 and len(testfiles) == 1:
        # if only one file is parsed, return the parseResults directly
        return retval[list(retval.keys())[0]]

    # else, return a dictionary of parseResults
    return retval

if __name__ == "__main__":
    main()



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ...._vendor._compat import basestring

from ..interface import Contract, ContractNotRespected, RValue, describe_value
from ..syntax import (W, oneOf, FollowedBy, NotAny)


class BindVariable(Contract):

    def __init__(self, variable, allowed_types, where=None):
        assert isinstance(variable, basestring) and len(variable) == 1
        assert allowed_types, '%r' % allowed_types
        Contract.__init__(self, where)
        self.variable = variable
        self.allowed_types = allowed_types

    def check_contract(self, context, value, silent):
        if self.variable in context:
            expected = context[self.variable]
            if not (expected == value):
                error = ('Expected value for %r was: %s\n        instead I received: %s' % (self.variable, describe_value(expected), describe_value(value)))
                raise ContractNotRespected(contract=self, error=error,
                                           value=value, context=context)

        else:
            # bound variable
            if not isinstance(value, self.allowed_types):
                error = ('Variable %r can only bind to %r, not %r.' %
                         (self.variable, self.allowed_types,
                          value.__class__.__name__))
                raise ContractNotRespected(self, error, value, context)

            context[self.variable] = value

    def __str__(self):
        return self.variable

    def __repr__(self):
        # XXX: invalid if tuple
        return 'BindVariable(%r,%s)' % (self.variable,
                                        self.allowed_types.__name__)

    @staticmethod
    def parse_action(allowed_types):
        def parse(s, loc, tokens):
            where = W(s, loc)
            variable = tokens[0]
            assert len(variable) == 1, \
                    ('Wrong syntax, matched %r as variable in %r.'
                     % (variable, s))
            # print ('Matched %r as variable in %r.' % (variable, s))
            return BindVariable(variable, allowed_types, where=where)
        return parse


class VariableRef(RValue):
    def __init__(self, variable, where=None):
        assert isinstance(variable, basestring)
        self.where = where
        self.variable = variable

    def eval(self, context):  # @ReservedAssignment
        var = self.variable
        if not var in context:
            raise ValueError('Unknown variable %r.' % var)
        return context[var]

    def __repr__(self):
        return "VariableRef(%r)" % self.variable

    def __str__(self):
        return "%s" % self.variable

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        return VariableRef(tokens[0], where=where)

alphabetu = 'A B C D E F G H I J K L M N O P Q R S T U W V X Y Z '
alphabetl = 'a b c d e f g h i j k l m n o p q r s t u w v x y z '

# Special case: allow an expression like AxBxC
nofollow = 'a b c d e f g h i j k l m n o p q r s t u w v   y z'
# also do not commit if part of word (SEn, a_2)
nofollow += ' A B C D E F G H I J K L M N O P Q R S T U W V X Y Z '
nofollow += ' 0 1 2 3 4 5 6 7 8 9 _'
# but recall 'axis_angle'
int_variables = (oneOf(alphabetu.split())
                  + FollowedBy(NotAny(oneOf(nofollow.split()))))
misc_variables = (oneOf(alphabetl.split())
                  + FollowedBy(NotAny(oneOf(nofollow.split() + ['x']))))
int_variables_ref = int_variables.copy().setParseAction(
                                                    VariableRef.parse_action)
misc_variables_ref = misc_variables.copy().setParseAction(
                                                    VariableRef.parse_action)

#int_variables = oneOf(alphabetu.split()) + FollowedBy(White() ^ 'x')

# These must be followed by whitespace; punctuation
#misc_variables = oneOf(alphabet.lower()) + FollowedBy(White()) 

nofollow = 'a b c d e f g h i j k l m n o p q r s t u w v   y z '
nofollow += ' * - + / '
nofollow += ' A B C D E F G H I J K L M N O P Q R S T U W V X Y Z '
nofollow += ' 0 1 2 3 4 5 6 7 8 9 _'
int_variables2 = (oneOf(alphabetu.split())
                  + FollowedBy(NotAny(oneOf(nofollow.split()))))
misc_variables2 = (oneOf(alphabetl.split())
                   + FollowedBy(NotAny(oneOf(nofollow.split() + ['x']))))
int_variables_contract = int_variables2.setParseAction(
                                                BindVariable.parse_action(int))
misc_variables_contract = misc_variables2.setParseAction(
                                            BindVariable.parse_action(object))





from pyparsing import *

wikiInput = """
Here is a simple Wiki input:
  *This is in italics.*
  **This is in bold!**
  ***This is in bold italics!***
  Here's a URL to {{Pyparsing's Wiki Page->https://site-closed.wikispaces.com}}
"""

def convertToHTML(opening,closing):
    def conversionParseAction(s,l,t):
        return opening + t[0] + closing
    return conversionParseAction

italicized = QuotedString("*").setParseAction(convertToHTML("<I>","</I>"))
bolded = QuotedString("**").setParseAction(convertToHTML("<B>","</B>"))
boldItalicized = QuotedString("***").setParseAction(convertToHTML("<B><I>","</I></B>"))
def convertToHTML_A(s,l,t):
    try:
        text,url=t[0].split("->")
    except ValueError:
        raise ParseFatalException(s,l,"invalid URL link reference: " + t[0])
    return '<A href="{0}">{1}</A>'.format(url, text)

urlRef = QuotedString("{{",endQuoteChar="}}").setParseAction(convertToHTML_A)

wikiMarkup = urlRef | boldItalicized | bolded | italicized

print(wikiInput)
print()
print(wikiMarkup.transformString(wikiInput))



import re
from itertools import chain
from typing import Union, Any, Optional

from ._util import invoke
from .core import MatchResult, MatchContext, transform, _, Underscore, Capture
from .error import MatchError
from .no_value import NoValue
from .patterns import InstanceOf, Regex
from .try_match import TryMatch


def _autopattern(pattern):
    if isinstance(pattern, type):
        return InstanceOf(pattern) & Underscore()
    if isinstance(pattern, re.Pattern):
        return Regex(pattern, capture_wildcards=True, bind_groups=False)
    if pattern is _:
        # this turns every underscore into a distinct object with it's own id - important
        # for recording wildcard matches via MatchContext.record
        return Underscore()
    return pattern


def match(value, pattern=NoValue, *extra,
          multimatch: bool = False,
          strict: bool = False,
          captureall: Optional[dict] = None) -> Union[MatchResult, Any]:
    """Matches the given value. Three different call styles are possible:

    (1) match(value, pattern, **kwargs)

    The so-called "simple" style which matches the given value against the given pattern, taking into account
    any options given as keyword-arguments and returns a MatchResult.

    (2) match(value)

    The "statement" style, to be used in a try...except block. This will actually raise a TryMatch Exception
    which is then caught by the matchers in the following except branches.

    (3) match(value, pattern1, action1, pattern2, action2, ...)

    The "terse" style. Every two arguments denote a pair of pattern and action. If pattern matches the value
    the corresponding action is invoked. If the action is actually callable it is called and it's return value
    is returned. It the action is a value then that value is simply returned.

    The "terse" style has a number of tweaks which are different from the matching in all other styles:

    - If a type is given as a pattern, the value is checked against that type using 'InstanceOf'. Each match
      is captured.
    - If a regular expression pattern ('re.Pattern', result of 're.compile') is given, then the value is checked
      against that regular expression and all capturing groups are captured.

    The captures are passed to the callable action.

    :param value: The value to be matched.
    :param pattern: The pattern to be matched against (if the "simple" style is used).
    :param multimatch: Whether to capture multiple matches per capture or keep the latest only (defaults to False).
    :param strict: Whether to perform strict matches (defaults to False).
    :param captureall: Capture all patterns into the given dictionary
    :return:
    """
    ctx = MatchContext(
        multimatch=multimatch,
        strict=strict,
    )
    if pattern is NoValue:
        raise TryMatch(value, ctx=ctx)
    elif extra:
        acc = []
        for p in chain((pattern,), extra):
            acc.append(transform(p, _autopattern))
            if len(acc) == 2:
                condition, action = acc
                result = match(value, condition)
                if result:
                    if callable(action):
                        return invoke(action, result.wildcard_matches())
                    return action
                acc = []
        if len(acc) == 1:
            if callable(acc[0]):
                return invoke(acc[0], [])
            return acc[0]
        raise MatchError(value)

    if isinstance(captureall, dict):
        count = 0

        def generate_name():
            nonlocal count
            count += 1
            return f"n{count}"

        pattern = transform(pattern, lambda x: Capture(x, name=generate_name(), target=captureall))

    result = ctx.match(value, pattern, strict=strict)
    return result



# jsonParser.py
#
# Copyright (c) 2006, Paul McGuire
#

test1 = """
{
        "glossary": {
            "title": "example glossary",
                "GlossDiv": {
                        "title": "S",
                        "GlossList": [{
                "ID": "SGML",
                "SortAs": "SGML",
                "GlossTerm": "Standard Generalized Markup Language",
                "Acronym": "SGML",
                "LargestPrimeLessThan100": 97,
                "AvogadroNumber": 6.02E23,
                "EvenPrimesGreaterThan2": null,
                "PrimesLessThan10" : [2,3,5,7],
                "WMDsFound" : false,
                "IraqAlQaedaConnections" : null,
                "Abbrev": "ISO 8879:1986",
                "GlossDef":
"A meta-markup language, used to create markup languages such as DocBook.",
                "GlossSeeAlso": ["GML", "XML", "markup"],
                "EmptyDict" : {},
                "EmptyList" : []
            }]
        }
    }
}
"""

test2 = """
{"menu": {
  "id": "file",
  "value": "File:",
  "popup": {
    "menuitem": [
      {"value": "New", "onclick": "CreateNewDoc()"},
      {"value": "Open", "onclick": "OpenDoc()"},
      {"value": "Close", "onclick": "CloseDoc()"}
    ]
  }
}}
"""
test3 = """
{"widget": {
    "debug": "on",
    "window": {
        "title": "Sample Konfabulator Widget",        "name": "main_window",        "width": 500,        "height": 500
    },    "image": {
        "src": "Images/Sun.png",
        "name": "sun1",        "hOffset": 250,        "vOffset": 250,        "alignment": "center"
    },    "text": {
        "data": "Click Here",
        "size": 36,
        "style": "bold",        "name": "text1",        "hOffset": 250,        "vOffset": 100,        "alignment": "center",
        "onMouseUp": "sun1.opacity = (sun1.opacity / 100) * 90;"
    }
}}
"""
test4 = """
{"web-app": {
  "servlet": [    // Defines the CDSServlet
    {
      "servlet-name": "cofaxCDS",
      "servlet-class": "org.cofax.cds.CDSServlet",
/*
    Defines glossary variables that template designers
    can use across the site.  You can add new
    variables to this set by creating a new init-param, with
    the param-name prefixed with "configGlossary:".
*/
      "init-param": {
        "configGlossary:installationAt": "Philadelphia, PA",
        "configGlossary:adminEmail": "ksm@pobox.com",
        "configGlossary:poweredBy": "Cofax",
        "configGlossary:poweredByIcon": "/images/cofax.gif",
        "configGlossary:staticPath": "/content/static",
/*
    Defines the template loader and template processor
    classes.  These are implementations of org.cofax.TemplateProcessor
    and org.cofax.TemplateLoader respectively.  Simply create new
    implementation of these classes and set them here if the default
    implementations do not suit your needs.  Leave these alone
    for the defaults.
*/
        "templateProcessorClass": "org.cofax.WysiwygTemplate",
        "templateLoaderClass": "org.cofax.FilesTemplateLoader",
        "templatePath": "templates",
        "templateOverridePath": "",
/*
    Defines the names of the default templates to look for
    when acquiring WYSIWYG templates.  Leave these at their
    defaults for most usage.
*/
        "defaultListTemplate": "listTemplate.htm",
        "defaultFileTemplate": "articleTemplate.htm",
/*
    New! useJSP switches on JSP template processing.
    jspListTemplate and jspFileTemplate are the names
    of the default templates to look for when aquiring JSP
    templates.  Cofax currently in production at KR has useJSP
    set to false, since our sites currently use WYSIWYG
    templating exclusively.
*/
        "useJSP": false,
        "jspListTemplate": "listTemplate.jsp",
        "jspFileTemplate": "articleTemplate.jsp",
/*
    Defines the packageTag cache.  This cache keeps
    Cofax from needing to interact with the database
    to look up packageTag commands.
*/
        "cachePackageTagsTrack": 200,
        "cachePackageTagsStore": 200,
        "cachePackageTagsRefresh": 60,
/*
    Defines the template cache.  Keeps Cofax from needing
    to go to the file system to load a raw template from
    the file system.
*/
        "cacheTemplatesTrack": 100,
        "cacheTemplatesStore": 50,
        "cacheTemplatesRefresh": 15,
/*
    Defines the page cache.  Keeps Cofax from processing
    templates to deliver to users.
*/
        "cachePagesTrack": 200,
        "cachePagesStore": 100,
        "cachePagesRefresh": 10,
        "cachePagesDirtyRead": 10,
/*
    Defines the templates Cofax will use when
    being browsed by a search engine identified in
    searchEngineRobotsDb
*/
        "searchEngineListTemplate": "forSearchEnginesList.htm",
        "searchEngineFileTemplate": "forSearchEngines.htm",
        "searchEngineRobotsDb": "WEB-INF/robots.db",
/*
    New!  useDataStore enables/disables the Cofax database pool
*/
        "useDataStore": true,
/*
    Defines the implementation of org.cofax.DataStore that Cofax
    will use.  If this DataStore class does not suit your needs
    simply implement a new DataStore class and set here.
*/
        "dataStoreClass": "org.cofax.SqlDataStore",
/*
    Defines the implementation of org.cofax.Redirection that
    Cofax will use.  If this Redirection class does not suit
    your needs simply implenet a new Redirection class
    and set here.
*/
        "redirectionClass": "org.cofax.SqlRedirection",
/*
    Defines the data store name.   Keep this at the default
*/
        "dataStoreName": "cofax",
/*
    Defines the JDBC driver that Cofax's database pool will use
*/
        "dataStoreDriver": "com.microsoft.jdbc.sqlserver.SQLServerDriver",
/*
    Defines the JDBC connection URL to connect to the database
*/
        "dataStoreUrl": "jdbc:microsoft:sqlserver://LOCALHOST:1433;DatabaseName=goon",
/*
    Defines the user name to connect to the database
*/
        "dataStoreUser": "sa",
/*
    Defines the password to connect to the database
*/
        "dataStorePassword": "dataStoreTestQuery",
/*
    A query that will run to test the validity of the
    connection in the pool.
*/
        "dataStoreTestQuery": "SET NOCOUNT ON;select test='test';",
/*
    A log file to print out database information
*/
        "dataStoreLogFile": "/usr/local/tomcat/logs/datastore.log",
/*
    The number of connection to initialize on startup
*/
        "dataStoreInitConns": 10,
/*
    The maximum number of connection to use in the pool
*/
        "dataStoreMaxConns": 100,
/*
    The number of times a connection will be utilized from the
    pool before disconnect
*/
        "dataStoreConnUsageLimit": 100,
/*
    The level of information to print to the log
*/
        "dataStoreLogLevel": "debug",
/*
    The maximum URL length allowable by the CDS Servlet
    Helps to prevent hacking
*/
        "maxUrlLength": 500}},
/*
    Defines the Email Servlet
*/
    {
      "servlet-name": "cofaxEmail",
      "servlet-class": "org.cofax.cds.EmailServlet",
      "init-param": {
/*
    The mail host to be used by the mail servlet
*/
        "mailHost": "mail1",
/*
    An override
*/
        "mailHostOverride": "mail2"}},
/*
    Defines the Admin Servlet - used to refresh cache on
    demand and see statistics
*/
    {
      "servlet-name": "cofaxAdmin",
      "servlet-class": "org.cofax.cds.AdminServlet"},
/*
    Defines the File Servlet - used to display files like Apache
*/
    {
      "servlet-name": "fileServlet",
      "servlet-class": "org.cofax.cds.FileServlet"},
    {
      "servlet-name": "cofaxTools",
      "servlet-class": "org.cofax.cms.CofaxToolsServlet",
      "init-param": {
/*
    Path to the template folder relative to the tools tomcat installation.
*/
        "templatePath": "toolstemplates/",
/*
    Logging boolean 1 = on, 0 = off
*/
        "log": 1,
/*
    Location of log. If empty, log will be written System.out
*/
        "logLocation": "/usr/local/tomcat/logs/CofaxTools.log",
/*
    Max size of log in BITS. If size is empty, no limit to log.
    If size is defined, log will be overwritten upon reaching defined size.
*/
        "logMaxSize": "",
/*
    DataStore logging boolean 1 = on, 0 = off
*/
        "dataLog": 1,
/*
    DataStore location of log. If empty, log will be written System.out
*/
        "dataLogLocation": "/usr/local/tomcat/logs/dataLog.log",
/*
    Max size of log in BITS. If size is empty, no limit to log.
    If size is defined, log will be overwritten upon reaching defined size.
*/
        "dataLogMaxSize": "",
/*
    Http string relative to server root to call for page cache
    removal to Cofax Servlet.
*/
        "removePageCache": "/content/admin/remove?cache=pages&id=",
/*
    Http string relative to server root to call for template
    cache removal to Cofax Servlet.
*/
        "removeTemplateCache": "/content/admin/remove?cache=templates&id=",
/*
    Location of folder from root of drive that will be used for
    ftp transfer from beta server or user hard drive to live servers.
    Note that Edit Article will not function without this variable
    set correctly. MultiPart request relies upon access to this folder.
*/
        "fileTransferFolder": "/usr/local/tomcat/webapps/content/fileTransferFolder",
/*
    Defines whether the Server should look in another path for
    config files or variables.
*/
        "lookInContext": 1,
/*
    Number of the ID of the top level administration group in tblPermGroups.
*/
        "adminGroupID": 4,
/*
    Is the tools app running on  the 'beta server'.
*/
        "betaServer": true}}],
  "servlet-mapping": {
/*
    URL mapping for the CDS Servlet
*/
     "cofaxCDS": "/",
/*
    URL mapping for the Email Servlet
*/
     "cofaxEmail": "/cofaxutil/aemail/*",
/*
    URL mapping for the Admin servlet
*/
     "cofaxAdmin": "/admin/*",
/*
    URL mapping for the Files servlet
*/
     "fileServlet": "/static/*",
     "cofaxTools": "/tools/*"},
/*
    New! The cofax taglib descriptor file
*/
  "taglib": {
    "taglib-uri": "cofax.tld",
    "taglib-location": "/WEB-INF/tlds/cofax.tld"}}}

"""

test5 = """
{"menu": {
    "header": "SVG Viewer",
    "items": [
        {"id": "Open"},
        {"id": "OpenNew", "label": "Open New"},
        null,
        {"id": "ZoomIn", "label": "Zoom In"},
        {"id": "ZoomOut", "label": "Zoom Out"},
        {"id": "OriginalView", "label": "Original View"},
        null,
        {"id": "Quality"},
        {"id": "Pause"},
        {"id": "Mute"},
        null,
        {"id": "Find", "label": "Find..."},
        {"id": "FindAgain", "label": "Find Again"},
        {"id": "Copy"},
        {"id": "CopyAgain", "label": "Copy Again"},
        {"id": "CopySVG", "label": "Copy SVG"},
        {"id": "ViewSVG", "label": "View SVG"},
        {"id": "ViewSource", "label": "View Source"},
        {"id": "SaveAs", "label": "Save As"},
        null,
        {"id": "Help"},
        {"id": "About", "label": "About Adobe CVG Viewer..."}
    ]
}}
"""



from __future__ import annotations

import unittest
from itertools import chain, combinations, permutations

from Aspidites._vendor.apm import *
from Aspidites._vendor.apm.core import transform


def powerset(it):
    s = list(it)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


def identity(x):
    return x


def record_type(s):
    def f(x):
        s.add(type(x))
        return x

    return f


class TransformTest(unittest.TestCase):

    def test_transform(self):
        ps0 = (
            Regex(r".+"),
            OneOf("foo", "bar"),
            AllOf("foo", "bar"),
            Either("foo", "bar"),
            Each(SubclassOf(int)),
            EachItem(InstanceOf(str), SubclassOf(type)),
            Parameters(int, k=str),
            Items(foo=17),
            Attrs(bar=117),
            At(path=["foo", "bar"], pattern=...),
            Arguments(foo=int, bar=float),
            Returns(str),
            [1, 2, 3],
            "foo",
            1.0,
        )
        ts = (
            Not,
            Strict,
            lambda pat: Transformed(identity, pat),
            lambda pat: Capture(pat, name="qux"),
        )
        ps = [
            AllOf(*ps0),
            OneOf(*ps0),
            NoneOf(*ps0),
        ]
        for p in ps0:
            for fs in powerset(ts):
                for fsp in permutations(fs):
                    pp = p
                    for f in fsp:
                        pp = f(pp)
                    ps.append(pp)

        for p in ps:
            types = set()
            t = transform(p, record_type(types))
            self.assertEqual(p, t)
            self.assertNotEqual(p, transform(p, lambda x: Not(x)))
            # since we already invested so much time into building combinations of patterns,
            # check that match does not error out (this will sometimes even match due to Not(...))
            expected_to_match = Not in types
            self.assertEqual(expected_to_match, bool(match(None, t)))
            # also do a check for captureall=, which is using a more fancy transform
            results = {}
            self.assertEqual(expected_to_match, bool(match(None, p, captureall=results)))
            if expected_to_match:
                self.assertTrue(bool(results))

    def test_some_transform(self):
        pt0 = (1, Many(...), 3)
        pt1 = transform(pt0, lambda x: x)
        self.assertEqual(pt0, pt1)

    def test_string_transform(self):
        pt0 = String("a", "b", "c")
        pt1 = transform(pt0, lambda x: x)
        self.assertEqual(pt0, pt1)

    def test_fancy_transform(self):
        # noinspection PyDefaultArgument
        def generate_name(static={'count': 0}):
            static['count'] += 1
            return f"n{static['count']}"

        pattern = (1, Many(...), 3)
        transformed = transform(pattern, lambda x: Capture(x, name=generate_name()))
        result = match((1, 2, 3), transformed)
        self.assertTrue(result)
        self.assertEqual({
            'n1': 1,
            'n2': 2,
            'n3': [2],
            'n4': 3,
            'n5': (1, 2, 3),
        }, result.groups())



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
import sys

from ..._compat import unicode
from ..interface import Contract, ContractNotRespected
from ..syntax import (add_contract, W, contract_expression, O, add_keyword,
    Keyword, Literal)


# Base class for string contracts
class StringBase(Contract):

    def __init__(self, length=None, where=None):
        Contract.__init__(self, where)
        self.length = length
        assert length is None or isinstance(length, Contract)

    def check_contract(self, context, value, silent):
        if not isinstance(value, self.TYPE):
            error = 'Expected %s, got %r.' % (
                self.DESCRIPTION, value.__class__.__name__)
            raise ContractNotRespected(contract=self, error=error,
                                       value=value, context=context)

        if self.length is not None:
            self.length._check_contract(context, len(value), silent)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.length)

    def __str__(self):
        s = self.KEYWORDS[0]
        if self.length is not None:
            s += '[%s]' % self.length
        return s

    @classmethod
    def parse_action(cls, s, loc, tokens):
        where = W(s, loc)
        length = tokens.get('length', None)
        return cls(length, where=where)


if sys.version_info[0] == 3:  # Python 3

    __all__ = ['String']

    class String(StringBase):
        KEYWORDS = ['str', 'string', 'unicode']
        TYPE = str
        DESCRIPTION = "a string"


else:  # Python 2.x

    __all__ = ['String', 'AnsiString', ]

    class String(StringBase):
        KEYWORDS = ['string']
        TYPE = basestring
        DESCRIPTION = "an ANSI or Unicode string"

    class AnsiString(StringBase):
        KEYWORDS = ['str']
        TYPE = str
        DESCRIPTION = "an ANSI string"


class UnicodeString(StringBase):
    KEYWORDS = ['unicode']
    TYPE = unicode
    DESCRIPTION = "a Unicode string"


__all__ += ['UnicodeString', ]

for cls in StringBase.__subclasses__():
    for keyword in cls.KEYWORDS:
        mycontract = (Keyword(keyword) +
                    O(Literal('[') - contract_expression('length') - ']'))
        add_keyword(keyword)
        add_contract(mycontract.setParseAction(cls.parse_action))



import os
import warnings
from pathlib import Path

import hypothesis
import pytest as pt
from hypothesis import given, assume, strategies as st

try:
    from numpy import inf, nan, isinf, isnan
except ImportError:
    from math import inf, nan, isinf, isnan


from .._vendor.contracts import ContractNotRespected
from ..__main__ import get_cy_kwargs, main
from ..parser import parse_module
from ..templates import woma_template
from ..monads import Maybe, Surely
from ..math import SafeFloorDiv, SafeMod, SafeDiv, SafeExp, Undefined
from ..compiler import Compiler, CompilerArgs

docker = os.getenv("ASPIDITES_DOCKER_BUILD")

woma_file = str(Path('examples/examples.wom'))


@pt.fixture(autouse=True)
def inject_config(request):
    return request.config.rootpath


def setup_code(inject_config):
    if Path(woma_file).exists():
        code_ = parse_module(open(woma_file, 'r').read())
    else:
        try:
            code_ = parse_module(open(str(Path("Aspidites/tests") / woma_file), 'r').read())
            warnings.warn("Aspidites is being tested in source mode")
        except FileNotFoundError:
            code_ = parse_module(open(Path(inject_config) / woma_file, 'r').read())
    return code_


def test_compile_module(inject_config):
    try:
        compile(woma_template.substitute(code='\n'.join(setup_code(inject_config))), '', 'exec')
    except FileNotFoundError:  # ????
        compile(woma_template.substitute(code='\n'.join(os.path.join(inject_config(), woma_file))), '',
                'exec')

def test_cli_help_exit():
    with pt.raises(SystemExit):
        main(['Aspidites', '-h'])


def test_cli_no_arg_exit():
    with pt.raises(SystemExit):
        main(['Aspidites' , ''])


def test_cli_no_target_exit():
    with pt.raises(SystemExit):
        main(['Aspidites', '-fpc'])


def test_cli_examples():
    python_file_ = Path('examples/compiled.py')
    python_file = python_file_ if Path(woma_file).exists() else (Path('Aspidites/tests') / python_file_)
    with pt.raises(SystemExit):
        main(['Aspidites', woma_file, f'-o={python_file}', '-c'])


# @pt.mark.filterwarnings('ignore::RuntimeWarning')
def test_compile_to_shared_object(inject_config):
    python_file_ = str(Path('examples/compiled.pyx'))
    full_path = Path('Aspidites/tests') / python_file_
    python_file = python_file_ if Path(woma_file).exists() else str(full_path)
    kwargs = get_cy_kwargs()
    code = setup_code(inject_config)
    kwargs.update(code=code, fname=python_file, bytecode=True, force=True, c=True, build_requires='', verbose=False)
    compile_args = CompilerArgs(**kwargs)
    try:
        Compiler(compile_args)
    except (FileNotFoundError, ValueError):
        kwargs.update(code=code, fname=Path(inject_config) / python_file_, bytecode=True)
        compile_args = CompilerArgs(**kwargs)
        Compiler(compile_args)

    from .examples.compiled import (Add, x, y, z, val, div_by_zero, Yield123, Hello, test_unary_sub, test_unary_add,
                                    test_factorial)


    with pt.raises(ContractNotRespected):
        # noinspection PyTypeChecker
        Add(x=6.5, y=12)

    assert [1, 2, 3] == [i for i in Yield123()]
    assert Maybe(Add, 6.5, 12)() == Undefined()
    assert x() == 6
    assert y() == Undefined()
    assert z == 9
    assert Add(x=3, y=2) == 5
    assert val() == Undefined()
    assert div_by_zero == Undefined()
    Hello()

#
# def teardown_function():
#     for file in glob.glob('examples/compiled.*'):
#         os.remove(file)
#     try:
#         os.remove('examples/setup.py')
#     except FileNotFoundError:
#         pass



from pyparsing import *

text = """Lorem ipsum dolor sit amet, consectetur adipisicing
elit, sed do eiusmod tempor incididunt ut labore et dolore magna
aliqua. Ut enim ad minim veniam, quis nostrud exercitation
ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis
aute irure dolor in reprehenderit in voluptate velit esse cillum
dolore eu fugiat nulla pariatur. Excepteur sint occaecat
cupidatat non proident, sunt in culpa qui officia deserunt
mollit anim id est laborum"""

# find all words beginning with a vowel
vowels = "aeiouAEIOU"
initialVowelWord = Word(vowels,alphas)

# Unfortunately, searchString will advance character by character through
# the input text, so it will detect that the initial "Lorem" is not an
# initialVowelWord, but then it will test "orem" and think that it is. So
# we need to add a do-nothing term that will match the words that start with
# consonants, but we will just throw them away when we match them. The key is
# that, in having been matched, the parser will skip over them entirely when
# looking for initialVowelWords.
consonants = ''.join(c for c in alphas if c not in vowels)
initialConsWord = Word(consonants, alphas).suppress()

# using scanString to locate where tokens are matched
for t,start,end in (initialConsWord|initialVowelWord).scanString(text):
    if t:
        print(start,':', t[0])

# add parse action to annotate the parsed tokens with their location in the
# input string
def addLocnToTokens(s,l,t):
    t['locn'] = l
    t['word'] = t[0]
initialVowelWord.setParseAction(addLocnToTokens)

for ivowelInfo in (initialConsWord | initialVowelWord).searchString(text):
    if not ivowelInfo:
        continue
    print(ivowelInfo.locn, ':', ivowelInfo.word)


# alternative - add an Empty that will save the current location
def location(name):
    return Empty().setParseAction(lambda s,l,t: t.__setitem__(name,l))
locateInitialVowels = location("locn") + initialVowelWord("word")

# search through the input text
for ivowelInfo in (initialConsWord | locateInitialVowels).searchString(text):
    if not ivowelInfo:
        continue
    print(ivowelInfo.locn, ':', ivowelInfo.word)



# rangeCheck.py
#
#   A sample program showing how parse actions can convert parsed
# strings into a data type or object, and to validate the parsed value.
#
# Updated to use new addCondition method and expr() copy.
#
# Copyright 2011,2015 Paul T. McGuire
#

from pyparsing import Word, nums, Suppress, Optional
from datetime import datetime

def ranged_value(expr, minval=None, maxval=None):
    # have to specify at least one range boundary
    if minval is None and maxval is None:
        raise ValueError("minval or maxval must be specified")

    # set range testing function and error message depending on
    # whether either or both min and max values are given
    inRangeCondition = {
        (True, False)  : lambda s,l,t : t[0] <= maxval,
        (False, True)  : lambda s,l,t : minval <= t[0],
        (False, False) : lambda s,l,t : minval <= t[0] <= maxval,
        }[minval is None, maxval is None]
    outOfRangeMessage = {
        (True, False)  : "value is greater than %s" % maxval,
        (False, True)  : "value is less than %s" % minval,
        (False, False) : "value is not in the range ({0} to {1})".format(minval,maxval),
        }[minval is None, maxval is None]

    return expr().addCondition(inRangeCondition, message=outOfRangeMessage)

# define the expressions for a date of the form YYYY/MM/DD or YYYY/MM (assumes YYYY/MM/01)
integer = Word(nums).setName("integer")
integer.setParseAction(lambda t:int(t[0]))

month = ranged_value(integer, 1, 12)
day = ranged_value(integer, 1, 31)
year = ranged_value(integer, 2000, None)

SLASH = Suppress('/')
dateExpr = year("year") + SLASH + month("month") + Optional(SLASH + day("day"))
dateExpr.setName("date")

# convert date fields to datetime (also validates dates as truly valid dates)
dateExpr.setParseAction(lambda t: datetime(t.year, t.month, t.day or 1).date())

# add range checking on dates
mindate = datetime(2002,1,1).date()
maxdate = datetime.now().date()
dateExpr = ranged_value(dateExpr, mindate, maxdate)


dateExpr.runTests("""
    2011/5/8
    2001/1/1
    2004/2/29
    2004/2
    1999/12/31""")



from __future__ import annotations

import unittest
from dataclasses import dataclass

from Aspidites._vendor.apm import *


@dataclass
class Record:
    foo: str
    bar: str


@dataclass
class User:
    foo: str
    bar: str


class UnhappyCasesTest(unittest.TestCase):

    def test_dataclass_no_match(self):
        self.assertFalse(match(Record("qux", "quuz"), 7))

    def test_dataclass_type_mismatch(self):
        self.assertFalse(match(Record("qux", "quuz"), User("qux", "quuz")))

    def test_strict_list_mismatch(self):
        xs = [0, 1, 2]
        vs = range(0, 3)
        self.assertTrue(match(vs, xs))
        self.assertFalse(match(vs, Strict(xs)))

    def test_simple_no_match(self):
        self.assertFalse(match(1, 2))

    def test_value_no_match(self):
        self.assertTrue(match(1, Value(1)))
        self.assertFalse(match(1, Value(2)))

    def test_strict_value_no_match(self):
        self.assertTrue(match(1, Value(1.0)))
        self.assertFalse(match(1, Strict(Value(1.0))))

    def test_invalid_values_for_some(self):
        with self.assertRaises(ValueError):
            Many(..., at_least=2, at_most=1)
        with self.assertRaises(ValueError):
            Many(..., at_least=2, exactly=1)
        with self.assertRaises(ValueError):
            Many(..., at_most=2, exactly=1)

    def test_invalid_values_for_length(self):
        with self.assertRaises(ValueError):
            Length(at_least=2, exactly=1)
        with self.assertRaises(ValueError):
            Length(at_most=2, exactly=1)

    def test_string_no_match(self):
        self.assertFalse(match("http://example.org/", String(
            Capture("http://", name='protocol'),
            Arguments(),
        )))

    def test_string_one_of_no_match(self):
        self.assertFalse(match("abc", String("ab", OneOf("e", "f"))))

    def test_string_regex_match(self):
        result = match("hello world", String("hello ", Regex("[A-Z][a-z]*")))
        self.assertFalse(result)

    def test_strict_value(self):
        self.assertTrue(match(3, Value(3.0)))
        self.assertTrue(match(3, Strict(Value(3))))
        self.assertFalse(match(3, Strict(Value(3.0))))

    def test_mismatching_exception(self):
        with self.assertRaises(TypeError):
            try:
                raise ValueError
            except Case(...):
                pass

    def test_parampattern_no_match(self):
        from Aspidites._vendor.apm.typefoo import ParamType, KwArgs, VarArgs
        self.assertFalse(match(1, KwArgs()))
        self.assertFalse(match(1, VarArgs()))
        self.assertFalse(match(1, ParamType(int)))



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8
import sys
from warnings import warn
from typing import Any, Union
from inspect import getouterframes
from math import inf, isinf, nan, isnan
from math import factorial
import numbers
import numpy as np
from .api import Warn

Numeric = Union[int, float, complex, np.number]


class Undefined:
    """A monad for a failed programmatic unit; like NoneType but hashable.
    Falsy singleton acts as an absorbing element for division."""

    __slots__ = ("__weakref__", "__instance__", "func", "args", "kwargs")

    def __hash__(self):
        # noinspection PyUnresolvedReferences
        return hash(self.__weakref__)

    def __eq__(self, other: Any):
        return type(other) == Undefined

    def __add__(self, other: Any):
        return self

    def __sub__(self, other: Any):
        return self

    def __mul__(self, other: Any):
        return self

    def __truediv__(self, other: Any):
        return self

    def __floordiv__(self, other):
        return self

    def __neg__(self):
        return self

    def __invert__(self):
        return self

    def __str__(self):
        return self.__repr__()

    def __int__(self):
        return nan

    def __float__(self):
        return nan

    def __complex__(self):
        return complex(nan)

    def __oct__(self):
        return self

    def __index__(self):
        return 0

    def __len__(self):
        # Undefined has 0 elements
        return 0

    def __repr__(self):
        if hasattr(self.func, '__name__'):
            r = self.__class__.__name__ + f"({self.func.__name__}, {self.args}, {self.kwargs})"
        else:
            r = self.__class__.__name__ + f"({None}, {self.args}, {self.kwargs})"
        return r

    # noinspection PyMethodMayBeStatic
    def __nonzero__(self):
        return True

    def __init__(self, func=None, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeFactorial(a: Numeric) -> Union[Numeric, Undefined]:
    if a < 0 or isnan(a) or isinf(a) or isinstance(a, (float, complex)):
        return Undefined(SafeFactorial, a)
    return factorial(a)


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeUnaryAdd(a: Numeric) -> Union[Numeric, Undefined]:
    if isnan(a) or not isinstance(a, numbers.Number):
        return Undefined(SafeUnaryAdd, a)
    return +a


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeUnarySub(a: Numeric) -> Union[Numeric, Undefined]:
    if isnan(a) or not isinstance(a, numbers.Number):
        return Undefined(SafeUnarySub, a)
    return -a


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeFloorDiv(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    if isinf(a) or b == 0 or (isinf(a) and isinf(b)):
        return Undefined(SafeFloorDiv, a, b)
    return a // b


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeDiv(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    if b == 0 or (isinf(a) and isinf(b)):
        return Undefined(SafeDiv, a, b)
    return a / b


# noinspection PyPep8Naming, PyProtectedMember,PyUnresolvedReferences
def SafeMod(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    if isinf(a) or b == 0:
        return Undefined(SafeMod, a, b)
    return a % b


# noinspection PyPep8Naming, PyProtectedMember,PyUnresolvedReferences
def SafeExp(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    if (a == 0 and b == 0) or (isinf(a) and b == 0) or (isinf(b) and a == 0):  # pragma: no cover
        return Undefined(SafeExp, a, b)
    try:
        return a ** b
    except OverflowError:
        return inf  # just a really big number on most systems



from __future__ import annotations

from ._util import invoke
from .match import match


class CaseExpr:
    def __init__(self, value, **kwargs):
        self._value = value
        self._kwargs = kwargs

    def of(self, pattern, then, **kwargs) -> CaseExpr:
        result = match(self._value, pattern, **{**self._kwargs, **kwargs})
        if result:
            if callable(then):
                return CaseExprEnd(invoke(then, result))
            else:
                return CaseExprEnd(then)
        else:
            return self

    def otherwise(self, then):
        if callable(then):
            return invoke(then, {})
        else:
            return then


class CaseExprEnd(CaseExpr):

    def of(self, pattern, then, **kwargs) -> CaseExpr:
        return self

    def otherwise(self, then):
        return self._value


def case(value) -> CaseExpr:
    return CaseExpr(value)



def _elements(thing):
    try:
        yield from thing.__dict__.values()
        return
    except AttributeError:
        pass
    try:
        yield from thing.values()
        return
    except AttributeError:
        pass
    try:
        yield from thing
        return
    except TypeError:
        pass
    yield thing
    return


def elements(thing):
    for x in _elements(thing):
        if isinstance(x, (dict, list)):
            yield tuple(elements(x))
        else:
            yield x


class AutoEqHash:

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __hash__(self):
        # noinspection PyTypeChecker
        return hash(tuple(elements(self)))


def _repr(t):
    if type(t) == dict:
        result = ["{"]
        one = False
        for k, v in t.items():
            if one:
                result.append(", ")
            if type(k) == str:
                result.append(k)
            else:
                result.append(repr(k))
            result.append("=")
            result.append(_repr(v))
            one = True
        result.append("}")
        return "".join(result)
    return repr(t)


class AutoRepr:

    def __repr__(self):
        return f"{type(self).__name__}({_repr(self.__dict__)})"



from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval
import warnings


def test_Is():
    assert restricted_eval('True is True') is True


def test_NotIs():
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        assert restricted_eval('1 is not True') is True



#
# A simple example showing the use of the implied listAllMatches=True for
# results names with a trailing '*' character.
#
# This example performs work similar to itertools.groupby, but without
# having to sort the input first.
#
# Copyright 2004-2016, by Paul McGuire
#
from ....._vendor.pyparsing import Word, ZeroOrMore, nums

aExpr = Word("A", nums)
bExpr = Word("B", nums)
cExpr = Word("C", nums)
grammar = ZeroOrMore(aExpr("A*") | bExpr("B*") | cExpr("C*"))

grammar.runTests("A1 B1 A2 C1 B2 A3")



'''
antlr_grammar.py

Created on 4 sept. 2010

@author: luca

Submitted by Luca DallOlio, September, 2010
(Minor updates by Paul McGuire, June, 2012)
(Code idiom updates by Paul McGuire, April, 2019)
'''
from ....._vendor.pyparsing import (Word, ZeroOrMore, printables, Suppress, OneOrMore, Group,
    LineEnd, Optional, White, originalTextFor, hexnums, nums, Combine, Literal, Keyword,
    cStyleComment, Regex, Forward, MatchFirst, And, oneOf, alphas, alphanums,
    delimitedList, Char)

# http://www.antlr.org/grammar/ANTLR/ANTLRv3.g

QUOTE,APOS,EQ,LBRACK,RBRACK,LBRACE,RBRACE,LPAR,RPAR,ROOT,BANG,AT,TIL,SEMI,COLON,VERT = map(Suppress,
                                                                                           '"\'=[]{}()^!@~;:|')
BSLASH = Literal('\\')
keywords = (SRC_, SCOPE_, OPTIONS_, TOKENS_, FRAGMENT, ID, LEXER, PARSER, GRAMMAR, TREE, CATCH, FINALLY,
            THROWS, PROTECTED, PUBLIC, PRIVATE, ) = map(Keyword,
    """src scope options tokens fragment id lexer parser grammar tree catch finally throws protected 
       public private """.split())
KEYWORD = MatchFirst(keywords)

# Tokens
EOL = Suppress(LineEnd()) # $
SGL_PRINTABLE = Char(printables)
singleTextString = originalTextFor(ZeroOrMore(~EOL + (White(" \t") | Word(printables)))).leaveWhitespace()
XDIGIT = hexnums
INT = Word(nums)
ESC = BSLASH + (oneOf(list(r'nrtbf\">'+"'")) | ('u' + Word(hexnums, exact=4)) | SGL_PRINTABLE)
LITERAL_CHAR = ESC | ~(APOS | BSLASH) + SGL_PRINTABLE
CHAR_LITERAL = APOS + LITERAL_CHAR + APOS
STRING_LITERAL = APOS + Combine(OneOrMore(LITERAL_CHAR)) + APOS
DOUBLE_QUOTE_STRING_LITERAL = '"' + ZeroOrMore(LITERAL_CHAR) + '"'
DOUBLE_ANGLE_STRING_LITERAL = '<<' + ZeroOrMore(SGL_PRINTABLE) + '>>'
TOKEN_REF = Word(alphas.upper(), alphanums+'_')
RULE_REF = Word(alphas.lower(), alphanums+'_')
ACTION_ESC = (BSLASH.suppress() + APOS
             | BSLASH.suppress()
             | BSLASH.suppress() + (~(APOS | QUOTE) + SGL_PRINTABLE)
              )
ACTION_CHAR_LITERAL = (APOS + (ACTION_ESC | ~(BSLASH | APOS) + SGL_PRINTABLE) + APOS)
ACTION_STRING_LITERAL = (QUOTE + ZeroOrMore(ACTION_ESC | ~(BSLASH | QUOTE) + SGL_PRINTABLE) + QUOTE)

SRC = SRC_.suppress() + ACTION_STRING_LITERAL("file") + INT("line")
id = TOKEN_REF | RULE_REF
SL_COMMENT = Suppress('//') + Suppress('$ANTLR') + SRC | ZeroOrMore(~EOL + Word(printables)) + EOL
ML_COMMENT = cStyleComment
WS = OneOrMore(Suppress(' ') | Suppress('\t') | (Optional(Suppress('\r')) + Literal('\n')))
WS_LOOP = ZeroOrMore(SL_COMMENT | ML_COMMENT)
NESTED_ARG_ACTION = Forward()
NESTED_ARG_ACTION << (LBRACK
                      + ZeroOrMore(NESTED_ARG_ACTION
                                   | ACTION_STRING_LITERAL
                                   | ACTION_CHAR_LITERAL)
                      + RBRACK)
ARG_ACTION = NESTED_ARG_ACTION
NESTED_ACTION = Forward()
NESTED_ACTION << (LBRACE
                  + ZeroOrMore(NESTED_ACTION
                               | SL_COMMENT
                               | ML_COMMENT
                               | ACTION_STRING_LITERAL
                               | ACTION_CHAR_LITERAL)
                  + RBRACE)
ACTION = NESTED_ACTION + Optional('?')
SCOPE = SCOPE_.suppress()
OPTIONS = OPTIONS_.suppress() + LBRACE # + WS_LOOP + Suppress('{')
TOKENS = TOKENS_.suppress() + LBRACE # + WS_LOOP + Suppress('{')
TREE_BEGIN = ROOT + LPAR
RANGE = Suppress('..')
REWRITE = Suppress('->')

# General Parser Definitions

# Grammar heading
optionValue = id | STRING_LITERAL | CHAR_LITERAL | INT | Literal('*').setName("s")

option = Group(id("id") + EQ + optionValue("value"))("option")
optionsSpec = OPTIONS + Group(OneOrMore(option + SEMI))("options") + RBRACE
tokenSpec = Group(TOKEN_REF("token_ref")
                  + (EQ + (STRING_LITERAL | CHAR_LITERAL)("lit")))("token") + SEMI
tokensSpec = TOKENS + Group(OneOrMore(tokenSpec))("tokens") + RBRACE
attrScope = SCOPE_.suppress() + id + ACTION
grammarType = LEXER + PARSER + TREE
actionScopeName = id | LEXER("l") | PARSER("p")
action = AT + Optional(actionScopeName + Suppress('::')) + id + ACTION

grammarHeading = (Optional(ML_COMMENT("ML_COMMENT"))
                  + Optional(grammarType)
                  + GRAMMAR
                  + id("grammarName") + SEMI
                  + Optional(optionsSpec)
                  + Optional(tokensSpec)
                  + ZeroOrMore(attrScope)
                  + ZeroOrMore(action))

modifier = PROTECTED | PUBLIC | PRIVATE | FRAGMENT
ruleAction = AT + id + ACTION
throwsSpec = THROWS.suppress() + delimitedList(id)
ruleScopeSpec = ((SCOPE_.suppress() + ACTION)
                 | (SCOPE_.suppress() + delimitedList(id) + SEMI)
                 | (SCOPE_.suppress() + ACTION + SCOPE_.suppress() + delimitedList(id) + SEMI))
unary_op = oneOf("^ !")
notTerminal = CHAR_LITERAL | TOKEN_REF | STRING_LITERAL
terminal = (CHAR_LITERAL | TOKEN_REF + Optional(ARG_ACTION) | STRING_LITERAL | '.') + Optional(unary_op)
block = Forward()
notSet = TIL + (notTerminal | block)
rangeNotPython = CHAR_LITERAL("c1") + RANGE + CHAR_LITERAL("c2")
atom = Group((rangeNotPython + Optional(unary_op)("op"))
             | terminal
             | (notSet + Optional(unary_op)("op"))
             | (RULE_REF + Optional(ARG_ACTION("arg")) + Optional(unary_op)("op"))
             )
element = Forward()
treeSpec = ROOT + LPAR + element*(2,) + RPAR
ebnfSuffix = oneOf("? * +")
ebnf = block + Optional(ebnfSuffix("op") | '=>')
elementNoOptionSpec = ((id("result_name")  + oneOf('= +=')("labelOp")  + atom("atom") + Optional(ebnfSuffix))
                       | (id("result_name") + oneOf('= +=')("labelOp") + block + Optional(ebnfSuffix))
                       | atom("atom") + Optional(ebnfSuffix)
                       | ebnf
                       | ACTION
                       | (treeSpec + Optional(ebnfSuffix))
                       ) # |   SEMPRED ( '=>' -> GATED_SEMPRED | -> SEMPRED )
element <<= Group(elementNoOptionSpec)("element")
# Do not ask me why group is needed twice... seems like the xml that you see is not always the real structure?
alternative = Group(Group(OneOrMore(element))("elements"))
rewrite = Optional(Literal('TODO REWRITE RULES TODO'))
block <<= (LPAR
           + Optional(Optional(optionsSpec("opts")) + COLON)
           + Group(alternative('a1')
                   + rewrite
                   + Group(ZeroOrMore(VERT
                                      + alternative('a2')
                                      + rewrite))("alternatives"))("block")
           + RPAR)
altList = alternative('a1') + rewrite + Group(ZeroOrMore(VERT + alternative('a2') + rewrite))("alternatives")
exceptionHandler = CATCH.suppress() + ARG_ACTION + ACTION
finallyClause = FINALLY.suppress() + ACTION
exceptionGroup = (OneOrMore(exceptionHandler) + Optional(finallyClause)) | finallyClause

ruleHeading = (Optional(ML_COMMENT)("ruleComment")
               + Optional(modifier)("modifier")
               + id("ruleName")
               + Optional("!")
               + Optional(ARG_ACTION("arg"))
               + Optional(Suppress('returns') + ARG_ACTION("rt"))
               + Optional(throwsSpec)
               + Optional(optionsSpec)
               + Optional(ruleScopeSpec)
               + ZeroOrMore(ruleAction))
rule = Group(ruleHeading + COLON + altList + SEMI + Optional(exceptionGroup))("rule")

grammarDef = grammarHeading + Group(OneOrMore(rule))("rules")

def grammar():
    return grammarDef

def __antlrAlternativesConverter(pyparsingRules, antlrBlock):
    rule = None
    if hasattr(antlrBlock, 'alternatives') and antlrBlock.alternatives != '' and len(antlrBlock.alternatives) > 0:
        alternatives = []
        alternatives.append(__antlrAlternativeConverter(pyparsingRules, antlrBlock.a1))
        for alternative in antlrBlock.alternatives:
            alternatives.append(__antlrAlternativeConverter(pyparsingRules, alternative))
        rule = MatchFirst(alternatives)("anonymous_or")
    elif hasattr(antlrBlock, 'a1') and antlrBlock.a1 != '':
        rule = __antlrAlternativeConverter(pyparsingRules, antlrBlock.a1)
    else:
        raise Exception('Not yet implemented')
    assert rule != None
    return rule

def __antlrAlternativeConverter(pyparsingRules, antlrAlternative):
    elementList = []
    for element in antlrAlternative.elements:
        rule = None
        if hasattr(element.atom, 'c1') and element.atom.c1 != '':
            regex = r'['+str(element.atom.c1[0])+'-'+str(element.atom.c2[0]+']')
            rule = Regex(regex)("anonymous_regex")
        elif hasattr(element, 'block') and element.block != '':
            rule = __antlrAlternativesConverter(pyparsingRules, element.block)
        else:
            ruleRef = element.atom[0]
            assert ruleRef in pyparsingRules
            rule = pyparsingRules[ruleRef](ruleRef)
        if hasattr(element, 'op') and element.op != '':
            if element.op == '+':
                rule = Group(OneOrMore(rule))("anonymous_one_or_more")
            elif element.op == '*':
                rule = Group(ZeroOrMore(rule))("anonymous_zero_or_more")
            elif element.op == '?':
                rule = Optional(rule)
            else:
                raise Exception('rule operator not yet implemented : ' + element.op)
        rule = rule
        elementList.append(rule)
    if len(elementList) > 1:
        rule = Group(And(elementList))("anonymous_and")
    else:
        rule = elementList[0]
    assert rule is not None
    return rule

def __antlrRuleConverter(pyparsingRules, antlrRule):
    rule = None
    rule = __antlrAlternativesConverter(pyparsingRules, antlrRule)
    assert rule != None
    rule(antlrRule.ruleName)
    return rule

def antlrConverter(antlrGrammarTree):
    pyparsingRules = {}

    antlrTokens = {}
    for antlrToken in antlrGrammarTree.tokens:
        antlrTokens[antlrToken.token_ref] = antlrToken.lit
    for antlrTokenName, antlrToken in list(antlrTokens.items()):
        pyparsingRules[antlrTokenName] = Literal(antlrToken)

    antlrRules = {}
    for antlrRule in antlrGrammarTree.rules:
        antlrRules[antlrRule.ruleName] = antlrRule
        pyparsingRules[antlrRule.ruleName] = Forward() # antlr is a top down grammar
    for antlrRuleName, antlrRule in list(antlrRules.items()):
        pyparsingRule = __antlrRuleConverter(pyparsingRules, antlrRule)
        assert pyparsingRule != None
        pyparsingRules[antlrRuleName] <<= pyparsingRule

    return pyparsingRules

if __name__ == "__main__":

    text = """\
grammar SimpleCalc;

options {
    language = Python;
}

tokens {
    PLUS     = '+' ;
    MINUS    = '-' ;
    MULT    = '*' ;
    DIV    = '/' ;
}

/*------------------------------------------------------------------
 * PARSER RULES
 *------------------------------------------------------------------*/

expr    : term ( ( PLUS | MINUS )  term )* ;

term    : factor ( ( MULT | DIV ) factor )* ;

factor    : NUMBER ;


/*------------------------------------------------------------------
 * LEXER RULES
 *------------------------------------------------------------------*/

NUMBER    : (DIGIT)+ ;

/* WHITESPACE : ( '\t' | ' ' | '\r' | '\n'| '\u000C' )+     { $channel = HIDDEN; } ; */

fragment DIGIT    : '0'..'9' ;

"""

    grammar().validate()
    antlrGrammarTree = grammar().parseString(text)
    print(antlrGrammarTree.dump())
    pyparsingRules = antlrConverter(antlrGrammarTree)
    pyparsingRule = pyparsingRules["expr"]
    pyparsingTree = pyparsingRule.parseString("2 - 5 * 42 + 7 / 25")
    print(pyparsingTree.dump())



from .  import syntax_fail, good, fail


good('map', {})
fail('map', 1)
fail('map', [])
syntax_fail('map[]')
syntax_fail('map[]()')
syntax_fail('map()')
good('map[1]', {1: 2})
good('map[N],N<2', {1: 2})
fail('map[N],N<2', {1: 2, 3: 4})
good('map(int:int)', {1: 2})
fail('map(int:int)', {'a': 2})
good('map(*:int)', {1: 2})
good('map(*:int)', {'a': 2})

# mapionary of string -> tuple, with tuple of two elements with different type
good('map(str:tuple(type(x),type(y))),x!=y', {'a': (2, 1.1)})
fail('map(str:tuple(type(x),type(y))),x!=y', {'a': (2, 1)})
